#!/usr/bin/env python
"""Test file for the physbiblio.gui.expWindows module.

This file is part of the physbiblio package.
"""
import os
import traceback
import unittest
from unittest.mock import MagicMock, call, patch

from PySide6.QtCore import QEvent, QModelIndex, Qt
from PySide6.QtGui import QMouseEvent
from PySide6.QtTest import QTest
from PySide6.QtWidgets import QWidget

try:
    from physbiblio.gui.commonClasses import PBTableView
    from physbiblio.gui.expWindows import *
    from physbiblio.gui.mainWindow import MainWindow
    from physbiblio.gui.setuptests import *
    from physbiblio.setuptests import *
except ImportError:
    print("Could not find physbiblio and its modules!")
    raise
except Exception:
    print(traceback.format_exc())


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class TestFunctions(GUIwMainWTestCase):
    """Test the editExperiment and deleteExperiment functions"""

    def test_editExperiment(self):
        """test editExperiment"""
        p = QWidget()
        m = self.mainW
        ncw = EditExperimentDialog(p)
        ncw.exec = MagicMock()
        ncw.onCancel()
        with patch(
            "physbiblio.gui.mainWindow.MainWindow.statusBarMessage", autospec=True
        ) as _s, patch(
            "physbiblio.gui.expWindows.EditExperimentDialog",
            return_value=ncw,
            autospec=USE_AUTOSPEC_CLASS,
        ) as _i:
            editExperiment(p, m)
            _i.assert_called_once_with(p, experiment=None)
            _s.assert_called_once_with(m, "No modifications to experiments")

        with patch(
            "physbiblio.gui.expWindows.EditExperimentDialog",
            return_value=ncw,
            autospec=USE_AUTOSPEC_CLASS,
        ) as _i, patch("logging.Logger.debug") as _l:
            editExperiment(p, p)
            _i.assert_called_once_with(p, experiment=None)
            _l.assert_called_once_with(
                "mainWinObject has no attribute 'statusBarMessage'", exc_info=True
            )

        with patch(
            "physbiblio.gui.mainWindow.MainWindow.statusBarMessage", autospec=True
        ) as _s, patch(
            "physbiblio.gui.expWindows.EditExperimentDialog",
            return_value=ncw,
            autospec=USE_AUTOSPEC_CLASS,
        ) as _i, patch(
            "physbiblio.database.Experiments.getDictByID",
            return_value="abc",
            autospec=True,
        ) as _g:
            editExperiment(p, m, 9999)
            _i.assert_called_once_with(p, experiment="abc")
            _g.assert_called_once_with(pBDB.exps, 9999)
            _s.assert_called_once_with(m, "No modifications to experiments")

        ncw = EditExperimentDialog(p)
        ncw.selectedExps = [9999]
        ncw.textValues["homepage"].setText("www.some.thing.com")
        ncw.textValues["comments"].setText("comm")
        ncw.textValues["inspire"].setText("1234")
        ncw.exec = MagicMock()
        ncw.onOk()
        with patch(
            "physbiblio.gui.mainWindow.MainWindow.statusBarMessage", autospec=True
        ) as _s, patch(
            "physbiblio.gui.expWindows.EditExperimentDialog",
            return_value=ncw,
            autospec=USE_AUTOSPEC_CLASS,
        ) as _i, patch(
            "physbiblio.database.Experiments.getDictByID",
            return_value="abc",
            autospec=True,
        ) as _g:
            editExperiment(p, m, editIdExp=9999)
            _i.assert_called_once_with(p, experiment="abc")
            _g.assert_called_once_with(pBDB.exps, 9999)
            _s.assert_called_once_with(m, "ERROR: empty experiment name")

        ncw.textValues["name"].setText("myexp")
        with patch(
            "physbiblio.gui.mainWindow.MainWindow.statusBarMessage", autospec=True
        ) as _s, patch(
            "physbiblio.gui.expWindows.EditExperimentDialog",
            return_value=ncw,
            autospec=USE_AUTOSPEC_CLASS,
        ) as _i, patch(
            "physbiblio.database.Experiments.getDictByID",
            return_value="abc",
            autospec=True,
        ) as _g, patch(
            "physbiblio.database.Experiments.insert", return_value="abc", autospec=True
        ) as _n, patch(
            "logging.Logger.debug"
        ) as _l:
            editExperiment(p, m, 9999)
            _i.assert_called_once_with(p, experiment="abc")
            _g.assert_called_once_with(pBDB.exps, 9999)
            _n.assert_called_once_with(
                pBDB.exps,
                {
                    "name": "myexp",
                    "inspire": "1234",
                    "homepage": "www.some.thing.com",
                    "comments": "comm",
                },
            )
            _s.assert_called_once_with(m, "Experiment saved")
            _l.assert_called_once_with(
                "parentObject has no attribute 'recreateTable'", exc_info=True
            )

        ncw.textValues["name"].setText("myexp")
        ctw = ExpsListWindow(p)
        with patch(
            "physbiblio.gui.mainWindow.MainWindow.statusBarMessage", autospec=True
        ) as _s, patch(
            "physbiblio.gui.expWindows.EditExperimentDialog",
            return_value=ncw,
            autospec=USE_AUTOSPEC_CLASS,
        ) as _i, patch(
            "physbiblio.database.Experiments.getDictByID",
            return_value="abc",
            autospec=True,
        ) as _g, patch(
            "physbiblio.database.Experiments.insert", return_value="abc", autospec=True
        ) as _n, patch(
            "logging.Logger.debug"
        ) as _l, patch(
            "physbiblio.gui.expWindows.ExpsListWindow.recreateTable", autospec=True
        ) as _r:
            editExperiment(ctw, m, 9999)
            _i.assert_called_once_with(ctw, experiment="abc")
            _g.assert_called_once_with(pBDB.exps, 9999)
            _n.assert_called_once_with(
                pBDB.exps,
                {
                    "name": "myexp",
                    "inspire": "1234",
                    "homepage": "www.some.thing.com",
                    "comments": "comm",
                },
            )
            _s.assert_called_once_with(m, "Experiment saved")
            _l.assert_not_called()
            _r.assert_called_once_with(ctw)

        exp = {
            "idExp": 9999,
            "homepage": "www.some.page.com",
            "comments": "no comment",
            "inspire": "1234",
            "name": "myexp",
        }
        ncw = EditExperimentDialog(p, exp)
        ncw.selectedCats = []
        ncw.textValues["inspire"].setText("4321")
        ncw.textValues["name"].setText("myexp1")
        ncw.textValues["comments"].setText("comm")
        ncw.textValues["homepage"].setText("www.page.com")
        ncw.exec = MagicMock()
        ncw.onOk()
        ctw = ExpsListWindow(p)
        with patch(
            "physbiblio.gui.mainWindow.MainWindow.statusBarMessage", autospec=True
        ) as _s, patch(
            "physbiblio.gui.expWindows.EditExperimentDialog",
            return_value=ncw,
            autospec=USE_AUTOSPEC_CLASS,
        ) as _i, patch(
            "physbiblio.database.Experiments.getDictByID",
            return_value="abc",
            autospec=True,
        ) as _g, patch(
            "physbiblio.database.Experiments.update", return_value="abc", autospec=True
        ) as _n, patch(
            "logging.Logger.info"
        ) as _l, patch(
            "physbiblio.gui.expWindows.ExpsListWindow.recreateTable", autospec=True
        ) as _r:
            editExperiment(ctw, m, 9999)
            _i.assert_called_once_with(ctw, experiment="abc")
            _g.assert_called_once_with(pBDB.exps, 9999)
            _n.assert_called_once_with(
                pBDB.exps,
                {
                    "idExp": "9999",
                    "name": "myexp1",
                    "homepage": "www.page.com",
                    "comments": "comm",
                    "inspire": "4321",
                },
                "9999",
            )
            _s.assert_called_once_with(m, "Experiment saved")
            _l.assert_called_once_with("Updating experiment 9999...")
            _r.assert_called_once_with(ctw)

    def test_deleteExperiment(self):
        """test deleteExperiment"""
        p = QWidget()
        m = self.mainW
        with patch(
            "physbiblio.gui.expWindows.askYesNo", return_value=False, autospec=True
        ) as _a, patch(
            "physbiblio.gui.mainWindow.MainWindow.statusBarMessage", autospec=True
        ) as _s:
            deleteExperiment(p, m, 9999, "myexp")
            _a.assert_called_once_with(
                "Do you really want to delete this experiment "
                + "(ID = '9999', name = 'myexp')?"
            )
            _s.assert_called_once_with(m, "Nothing changed")

        with patch(
            "physbiblio.gui.expWindows.askYesNo", return_value=False, autospec=True
        ) as _a, patch("logging.Logger.debug") as _d:
            deleteExperiment(p, p, 9999, "myexp")
            _a.assert_called_once_with(
                "Do you really want to delete this experiment "
                + "(ID = '9999', name = 'myexp')?"
            )
            _d.assert_called_once_with(
                "mainWinObject has no attribute 'statusBarMessage'", exc_info=True
            )

        with patch(
            "physbiblio.gui.expWindows.askYesNo", return_value=True, autospec=True
        ) as _a, patch(
            "physbiblio.database.Experiments.delete", autospec=True
        ) as _c, patch(
            "PySide6.QtWidgets.QMainWindow.setWindowTitle", autospec=True
        ) as _t, patch(
            "physbiblio.gui.mainWindow.MainWindow.statusBarMessage", autospec=True
        ) as _s, patch(
            "logging.Logger.debug"
        ) as _d:
            deleteExperiment(p, m, 9999, "myexp")
            _a.assert_called_once_with(
                "Do you really want to delete this experiment "
                + "(ID = '9999', name = 'myexp')?"
            )
            _c.assert_called_once_with(pBDB.exps, 9999)
            _t.assert_called_once_with("PhysBiblio*")
            _s.assert_called_once_with(m, "Experiment deleted")
            _d.assert_called_once_with(
                "parentObject has no attribute 'recreateTable'", exc_info=True
            )

        elw = ExpsListWindow(p)
        with patch(
            "physbiblio.gui.expWindows.askYesNo", return_value=True, autospec=True
        ) as _a, patch(
            "physbiblio.database.Experiments.delete", autospec=True
        ) as _c, patch(
            "PySide6.QtWidgets.QMainWindow.setWindowTitle", autospec=True
        ) as _t, patch(
            "physbiblio.gui.mainWindow.MainWindow.statusBarMessage", autospec=True
        ) as _s, patch(
            "physbiblio.gui.expWindows.ExpsListWindow.recreateTable", autospec=True
        ) as _r:
            deleteExperiment(elw, m, 9999, "myexp")
            _a.assert_called_once_with(
                "Do you really want to delete this experiment "
                + "(ID = '9999', name = 'myexp')?"
            )
            _c.assert_called_once_with(pBDB.exps, 9999)
            _t.assert_called_once_with("PhysBiblio*")
            _s.assert_called_once_with(m, "Experiment deleted")
            _r.assert_called_once_with(elw)


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class TestExpTableModel(GUITestCase):
    """Test ExpTableModel"""

    def test_init(self):
        """test init"""
        p = QWidget()
        exp_list = [{"idExp": 0, "name": "no"}]
        header = ["id", "name"]
        em = ExpTableModel(p, exp_list, header, askExps=False, previous=[])
        self.assertIsInstance(em, PBTableModel)
        self.assertEqual(em.typeClass, "Exps")
        self.assertEqual(em.dataList, exp_list)
        with patch(
            "physbiblio.gui.commonClasses.PBTableModel.__init__", autospec=True
        ) as _i, patch(
            "physbiblio.gui.commonClasses.PBTableModel.prepareSelected", autospec=True
        ) as _s:
            em = ExpTableModel(p, exp_list, header, askExps=True, previous=[9999])
            _i.assert_called_once_with(em, p, ["id", "name"], True, [9999])
            _s.assert_called_once_with(em)

    def test_getIdentifier(self):
        """test getIdentifier"""
        p = QWidget()
        exp_list = [{"idExp": 0, "name": "no"}]
        header = ["id", "name"]
        em = ExpTableModel(p, exp_list, header, askExps=False, previous=[])
        self.assertEqual(em.getIdentifier({"idExp": 9999, "name": "no"}), 9999)

    def test_data(self):
        """test data"""
        p = QWidget()
        exp_list = [{"idExp": 0, "name": "myexp"}, {"idExp": 9999, "name": "exp"}]
        header = ["idExp", "name"]
        em = ExpTableModel(p, exp_list, header, askExps=False, previous=[])
        ix = em.index(10, 0)
        self.assertEqual(em.data(ix, Qt.CheckStateRole), None)
        ix = em.index(0, 0)
        self.assertEqual(em.data(ix, Qt.DisplayRole), 0)
        self.assertEqual(em.data(ix, Qt.EditRole), 0)
        self.assertEqual(em.data(ix, Qt.DecorationRole), None)
        ix = em.index(0, 1)
        self.assertEqual(em.data(ix, Qt.DisplayRole), "myexp")
        self.assertEqual(em.data(ix, Qt.EditRole), "myexp")
        self.assertEqual(em.data(ix, Qt.DecorationRole), None)

        ix = em.index(0, 0)
        self.assertEqual(em.data(ix, Qt.CheckStateRole), None)
        p = QWidget()
        p.askExps = False
        em = ExpTableModel(p, exp_list, header, askExps=False, previous=[])
        self.assertEqual(em.data(em.index(0, 0), Qt.CheckStateRole), None)
        em = ExpTableModel(p, exp_list, header, askExps=True, previous=[])
        self.assertEqual(em.data(em.index(0, 0), Qt.CheckStateRole), Qt.Unchecked)
        em = ExpTableModel(p, exp_list, header, askExps=True, previous=[0])
        self.assertEqual(em.data(em.index(0, 0), Qt.CheckStateRole), Qt.Checked)
        with self.assertRaises(ValueError):
            em.setData(ix, "abc", Qt.CheckStateRole)
        self.assertEqual(em.data(em.index(0, 0), Qt.CheckStateRole), Qt.Checked)
        self.assertEqual(em.setData(ix, Qt.Unchecked, Qt.CheckStateRole), True)
        self.assertEqual(em.data(em.index(0, 0), Qt.CheckStateRole), Qt.Unchecked)

    def test_setData(self):
        """test setData"""
        p = QWidget()

        def connectEmit(ix1, ix2):
            """used to test dataChanged.emit"""
            self.newEmit = ix1

        exp_list = [{"idExp": 0, "name": "no"}, {"idExp": 9999, "name": "yes"}]
        header = ["idExp", "name"]
        cm = ExpTableModel(p, exp_list, header)
        ix = cm.index(10, 0)
        self.newEmit = False
        cm.dataChanged.connect(connectEmit)
        self.assertEqual(cm.setData(ix, "abc", Qt.CheckStateRole), False)
        self.assertEqual(self.newEmit, False)
        ix = cm.index(0, 0)
        self.assertEqual(cm.setData(ix, Qt.Checked, Qt.CheckStateRole), True)
        self.assertEqual(cm.selectedElements[0], True)
        self.assertEqual(self.newEmit, ix)
        self.newEmit = False
        with self.assertRaises(ValueError):
            cm.setData(ix, "abc", Qt.CheckStateRole)
        self.assertEqual(cm.selectedElements[0], True)
        self.assertEqual(self.newEmit, False)

        cm = ExpTableModel(p, exp_list, header)
        ix = cm.index(0, 0)
        self.assertEqual(cm.setData(ix, "abc", Qt.EditRole), True)
        self.assertEqual(cm.selectedElements[0], False)


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class TestExpsListWindow(GUITestCase):
    """Test ExpsListWindow"""

    def setUp(self):
        """define common parameters for test use"""
        self.exps = [
            {
                "idExp": 0,
                "name": "test0",
                "homepage": "",
                "comments": "",
                "inspire": "",
            },
            {
                "idExp": 1,
                "name": "test1",
                "homepage": "",
                "comments": "",
                "inspire": "1234",
            },
            {
                "idExp": 2,
                "name": "test2",
                "homepage": "",
                "comments": "",
                "inspire": "",
            },
            {
                "idExp": 3,
                "name": "test3",
                "homepage": "",
                "comments": "",
                "inspire": "",
            },
        ]

    def test_init(self):
        """test init"""
        p = QWidget()
        with patch(
            "physbiblio.database.Experiments.getAll",
            return_value=self.exps,
            autospec=True,
        ) as _gh, patch(
            "physbiblio.gui.expWindows.ExpsListWindow.createTable", autospec=True
        ) as _cf:
            elw = ExpsListWindow(p)
            _cf.assert_called_once_with(elw)
        self.assertIsInstance(elw, ObjListWindow)
        self.assertEqual(elw.parent(), p)
        self.assertEqual(elw.colcnt, 5)
        self.assertEqual(
            elw.colContents, ["idExp", "name", "comments", "homepage", "inspire"]
        )
        self.assertEqual(elw.askExps, False)
        self.assertEqual(elw.askForBib, None)
        self.assertEqual(elw.askForCat, None)
        self.assertEqual(elw.previous, [])
        self.assertEqual(elw.windowTitle(), "List of experiments")

        with patch(
            "physbiblio.database.Experiments.getAll",
            return_value=self.exps,
            autospec=True,
        ) as _gh, patch(
            "physbiblio.gui.expWindows.ExpsListWindow.createTable", autospec=True
        ) as _cf:
            elw = ExpsListWindow(
                parent=p,
                askExps=True,
                askForBib="mybib",
                askForCat="mycat",
                previous=[9999],
            )
            _cf.assert_called_once_with(elw)
        self.assertIsInstance(elw, ObjListWindow)
        self.assertEqual(elw.parent(), p)
        self.assertEqual(elw.colcnt, 5)
        self.assertEqual(
            elw.colContents, ["idExp", "name", "comments", "homepage", "inspire"]
        )
        self.assertEqual(elw.askExps, True)
        self.assertEqual(elw.askForBib, "mybib")
        self.assertEqual(elw.askForCat, "mycat")
        self.assertEqual(elw.previous, [9999])
        self.assertEqual(elw.windowTitle(), "List of experiments")

    def test_populateAskExp(self):
        """test populateAskExp"""
        p = QWidget()
        with patch(
            "physbiblio.database.Experiments.getAll",
            return_value=self.exps,
            autospec=True,
        ) as _gh, patch(
            "physbiblio.gui.expWindows.ExpsListWindow.createTable", autospec=True
        ) as _c:
            elw = ExpsListWindow(p)
        # test with askCats = False (nothing happens)
        self.assertTrue(elw.populateAskExp())
        self.assertEqual(elw.marked, [])
        self.assertFalse(hasattr(p, "selectedExps"))
        self.assertEqual(elw.currLayout.count(), 0)

        # test with askCats = True and askForBib
        with patch(
            "physbiblio.database.Experiments.getAll",
            return_value=self.exps,
            autospec=True,
        ) as _gh, patch(
            "physbiblio.gui.expWindows.ExpsListWindow.createTable", autospec=True
        ) as _c:
            elw = ExpsListWindow(p, askExps=True, askForBib="bib")
        # bibkey not in db
        with patch(
            "physbiblio.database.Entries.getByBibkey", return_value=[], autospec=True
        ) as _gbb, patch("logging.Logger.warning") as _w:
            self.assertEqual(elw.populateAskExp(), None)
            _gbb.assert_called_once_with(pBDB.bibs, "bib", saveQuery=False)
            _w.assert_called_once_with(
                "The entry 'bib' is not in the database!", exc_info=True
            )
        self.assertEqual(elw.marked, [])
        self.assertFalse(hasattr(p, "selectedExps"))
        self.assertEqual(elw.currLayout.count(), 0)
        # dict has no "title" or "author"
        with patch(
            "physbiblio.database.Entries.getByBibkey",
            side_effect=[[{"inspire": None, "doi": None, "arxiv": None}]],
            autospec=True,
        ) as _gbb:
            self.assertEqual(elw.populateAskExp(), True)
            _gbb.assert_called_once_with(pBDB.bibs, "bib", saveQuery=False)
            self.assertEqual(elw.currLayout.count(), 1)
            self.assertIsInstance(elw.layout().itemAt(0).widget(), PBLabel)
            self.assertEqual(
                elw.layout().itemAt(0).widget().text(),
                "Mark experiments for the following "
                + "entry:<br><b>key</b>:<br>bib<br>",
            )
        self.assertEqual(elw.marked, [])
        self.assertTrue(hasattr(p, "selectedExps"))
        self.assertEqual(p.selectedExps, [])
        # no link exists
        elw.cleanLayout()
        with patch(
            "physbiblio.database.Entries.getByBibkey",
            side_effect=[
                [
                    {
                        "author": "sg",
                        "title": "title",
                        "inspire": None,
                        "doi": None,
                        "arxiv": None,
                    }
                ]
            ],
            autospec=True,
        ) as _gbb:
            self.assertEqual(elw.populateAskExp(), True)
            _gbb.assert_called_once_with(pBDB.bibs, "bib", saveQuery=False)
            self.assertEqual(elw.currLayout.count(), 1)
            self.assertIsInstance(elw.layout().itemAt(0).widget(), PBLabel)
            self.assertEqual(
                elw.layout().itemAt(0).widget().text(),
                "Mark experiments for the following "
                + "entry:<br><b>key</b>:<br>bib<br>"
                + "<b>author(s)</b>:<br>sg<br>"
                + "<b>title</b>:<br>title<br>",
            )
        # inspire exists
        elw.cleanLayout()
        with patch(
            "physbiblio.database.Entries.getByBibkey",
            side_effect=[
                [
                    {
                        "author": "sg",
                        "title": "title",
                        "inspire": "1234",
                        "doi": "1/2/3",
                        "arxiv": "123456",
                    }
                ]
            ],
            autospec=True,
        ) as _gbb, patch(
            "physbiblio.view.ViewEntry.getLink",
            return_value="inspirelink",
            autospec=True,
        ) as _l:
            self.assertEqual(elw.populateAskExp(), True)
            _gbb.assert_called_once_with(pBDB.bibs, "bib", saveQuery=False)
            _l.assert_called_once_with(pBView, "bib", "inspire")
            self.assertEqual(elw.currLayout.count(), 1)
            self.assertIsInstance(elw.layout().itemAt(0).widget(), PBLabel)
            self.assertEqual(
                elw.layout().itemAt(0).widget().text(),
                "Mark experiments for the following "
                + "entry:<br><b>key</b>:<br><a href='inspirelink'>bib</a><br>"
                + "<b>author(s)</b>:<br>sg<br>"
                + "<b>title</b>:<br>title<br>",
            )
        elw.cleanLayout()
        # arxiv exists
        with patch(
            "physbiblio.database.Entries.getByBibkey",
            side_effect=[
                [
                    {
                        "author": "sg",
                        "title": "title",
                        "inspire": None,
                        "doi": "1/2/3",
                        "arxiv": "123456",
                    }
                ]
            ],
            autospec=True,
        ) as _gbb, patch(
            "physbiblio.view.ViewEntry.getLink", return_value="arxivlink", autospec=True
        ) as _l:
            self.assertEqual(elw.populateAskExp(), True)
            _gbb.assert_called_once_with(pBDB.bibs, "bib", saveQuery=False)
            _l.assert_called_once_with(pBView, "bib", "arxiv")
            self.assertEqual(elw.currLayout.count(), 1)
            self.assertIsInstance(elw.layout().itemAt(0).widget(), PBLabel)
            self.assertEqual(
                elw.layout().itemAt(0).widget().text(),
                "Mark experiments for the following "
                + "entry:<br><b>key</b>:<br><a href='arxivlink'>bib</a><br>"
                + "<b>author(s)</b>:<br>sg<br>"
                + "<b>title</b>:<br>title<br>",
            )
        elw.cleanLayout()
        # doi exists
        with patch(
            "physbiblio.database.Entries.getByBibkey",
            side_effect=[
                [
                    {
                        "author": "sg",
                        "title": "title",
                        "inspire": None,
                        "doi": "1/2/3",
                        "arxiv": "",
                    }
                ]
            ],
            autospec=True,
        ) as _gbb, patch(
            "physbiblio.view.ViewEntry.getLink", return_value="doilink", autospec=True
        ) as _l:
            self.assertEqual(elw.populateAskExp(), True)
            _gbb.assert_called_once_with(pBDB.bibs, "bib", saveQuery=False)
            _l.assert_called_once_with(pBView, "bib", "doi")
            self.assertEqual(elw.currLayout.count(), 1)
            self.assertIsInstance(elw.layout().itemAt(0).widget(), PBLabel)
            self.assertEqual(
                elw.layout().itemAt(0).widget().text(),
                "Mark experiments for the following "
                + "entry:<br><b>key</b>:<br><a href='doilink'>bib</a><br>"
                + "<b>author(s)</b>:<br>sg<br>"
                + "<b>title</b>:<br>title<br>",
            )

        # test with askCats = True and askForExp
        p = QWidget()
        with patch(
            "physbiblio.database.Experiments.getAll",
            return_value=self.exps,
            autospec=True,
        ) as _gh, patch(
            "physbiblio.gui.expWindows.ExpsListWindow.createTable", autospec=True
        ) as _c:
            elw = ExpsListWindow(p, askExps=True, askForCat="cat")
            with self.assertRaises(NotImplementedError):
                elw.populateAskExp()

        # #test with no askForBib, askForExp
        p = QWidget()
        with patch(
            "physbiblio.database.Experiments.getAll",
            return_value=self.exps,
            autospec=True,
        ) as _gh, patch(
            "physbiblio.gui.expWindows.ExpsListWindow.createTable", autospec=True
        ) as _c:
            elw = ExpsListWindow(p, askExps=True)
        self.assertEqual(elw.populateAskExp(), True)
        self.assertEqual(elw.currLayout.count(), 1)
        self.assertIsInstance(elw.layout().itemAt(0).widget(), PBLabel)
        self.assertEqual(
            elw.layout().itemAt(0).widget().text(), "Select the desired experiments:"
        )

    def test_onCancel(self):
        """test onCancel"""
        elw = ExpsListWindow()
        with patch("PySide6.QtWidgets.QDialog.close", autospec=True) as _c:
            elw.onCancel()
        self.assertFalse(elw.result)

    def test_onOk(self):
        """test onOk"""
        p = QWidget()
        with patch(
            "physbiblio.database.Experiments.getAll",
            return_value=self.exps,
            autospec=True,
        ) as _gh:
            elw = ExpsListWindow(p)
        self.assertFalse(hasattr(p, "selectedExps"))
        elw.tableModel.selectedElements[0] = True
        elw.tableModel.selectedElements[2] = False
        elw.tableModel.selectedElements[3] = True
        with patch("PySide6.QtWidgets.QDialog.close", autospec=True) as _c:
            elw.onOk()
            _c.assert_called_once_with()
        self.assertEqual(hasattr(p, "selectedExps"), True)
        self.assertEqual(p.selectedExps, [0, 3])
        self.assertEqual(elw.result, "Ok")

    def test_onNewExp(self):
        """test onNewExp"""
        p = QWidget()
        elw = ExpsListWindow(p)
        with patch("physbiblio.gui.expWindows.editExperiment", autospec=True) as _ec:
            elw.onNewExp()
            _ec.assert_called_once_with(elw, p)

    def test_keyPressEvent(self):
        """test keyPressEvent"""
        elw = ExpsListWindow()
        elw.filterInput.setFocus = MagicMock()
        with patch("PySide6.QtWidgets.QDialog.close", autospec=True) as _oc, patch(
            "physbiblio.gui.expWindows.ExpsListWindow.onOk", autospec=True
        ) as _ok:
            QTest.keyPress(elw, "a")
            _oc.assert_not_called()
            QTest.keyPress(elw, Qt.Key_Enter)
            _oc.assert_not_called()
            QTest.keyPress(elw, Qt.Key_Escape)
            self.assertEqual(_oc.call_count, 1)
            QTest.keyPress(elw, "a")
            _ok.assert_not_called()
            QTest.keyPress(elw, Qt.Key_Enter)
            _ok.assert_not_called()
            QTest.keyPress(elw, Qt.Key_Escape)
            _ok.assert_not_called()
            QTest.keyPress(elw, Qt.Key_Return, Qt.ControlModifier)
            self.assertEqual(_ok.call_count, 1)
            QTest.keyPress(elw, Qt.Key_Enter, Qt.ControlModifier)
            self.assertEqual(_ok.call_count, 2)
            QTest.keyPress(elw, Qt.Key_F)
            elw.filterInput.setFocus.assert_not_called()
            QTest.keyPress(elw, Qt.Key_F, Qt.ControlModifier)
            elw.filterInput.setFocus.assert_called_once_with()

    def test_createTable(self):
        """test createTable"""
        p = QWidget()
        with patch(
            "physbiblio.gui.expWindows.ExpsListWindow.createTable", autospec=True
        ) as _c:
            elw = ExpsListWindow(p)
        with patch(
            "physbiblio.database.Experiments.getAll",
            return_value=self.exps,
            autospec=True,
        ) as _gh, patch(
            "physbiblio.gui.expWindows.ExpsListWindow.populateAskExp", autospec=True
        ) as _pae, patch(
            "physbiblio.gui.expWindows.ExpTableModel.__init__",
            return_value=None,
            autospec=True,
        ) as _etm, patch(
            "physbiblio.gui.commonClasses.ObjListWindow.addFilterInput", autospec=True
        ) as _afi, patch(
            "physbiblio.gui.commonClasses.ObjListWindow.setProxyStuff", autospec=True
        ) as _sps, patch(
            "physbiblio.gui.commonClasses.ObjListWindow.finalizeTable", autospec=True
        ) as _ft:
            elw.createTable()
            _pae.assert_called_once_with(elw)
            _gh.assert_called_once_with(pBDB.exps)
            _afi.assert_called_once_with(elw, "Filter experiment")
            _sps.assert_called_once_with(elw, 1, Qt.AscendingOrder)
            _ft.assert_called_once_with(elw)
            _etm.assert_called_once_with(
                elw.tableModel,
                elw,
                self.exps,
                pBDB.tableCols["experiments"],
                askExps=False,
                previous=[],
            )
        elw.cleanLayout()
        with patch(
            "physbiblio.database.Experiments.getAll",
            return_value=self.exps,
            autospec=True,
        ) as _gh:
            elw.createTable()
        self.assertEqual(self.exps, elw.exps)
        self.assertIsInstance(elw.tableModel, ExpTableModel)
        self.assertEqual(elw.layout().count(), 3)
        self.assertIsInstance(elw.layout().itemAt(0).widget(), QLineEdit)
        self.assertEqual(elw.layout().itemAt(0).widget(), elw.filterInput)
        self.assertEqual(elw.filterInput.placeholderText(), "Filter experiment")
        with patch(
            "physbiblio.gui.expWindows.ExpsListWindow.changeFilter", autospec=True
        ) as _cf:
            elw.filterInput.textChanged.emit("a")
            _cf.assert_called_once_with(elw, "a")
        self.assertIsInstance(elw.layout().itemAt(1).widget(), PBTableView)
        self.assertEqual(elw.layout().itemAt(1).widget(), elw.tableview)
        self.assertIsInstance(elw.layout().itemAt(2).widget(), QPushButton)
        self.assertEqual(elw.layout().itemAt(2).widget(), elw.newExpButton)
        self.assertEqual(elw.newExpButton.text(), "Add new experiment")
        with patch(
            "physbiblio.gui.expWindows.ExpsListWindow.onNewExp", autospec=True
        ) as _f:
            QTest.mouseClick(elw.newExpButton, Qt.LeftButton)
            _f.assert_called_once_with(elw)

        # repeat with askExps=True, but ignore populateAskExp
        with patch(
            "physbiblio.gui.expWindows.ExpsListWindow.populateAskExp", autospec=True
        ) as _c:
            elw = ExpsListWindow(p, askExps=True)
        self.assertEqual(elw.layout().count(), 5)
        self.assertIsInstance(elw.layout().itemAt(0).widget(), QLineEdit)
        self.assertEqual(elw.layout().itemAt(0).widget(), elw.filterInput)
        self.assertEqual(elw.filterInput.placeholderText(), "Filter experiment")
        with patch(
            "physbiblio.gui.expWindows.ExpsListWindow.changeFilter", autospec=True
        ) as _cf:
            elw.filterInput.textChanged.emit("a")
            _cf.assert_called_once_with(elw, "a")
        self.assertIsInstance(elw.layout().itemAt(1).widget(), PBTableView)
        self.assertEqual(elw.layout().itemAt(1).widget(), elw.tableview)
        self.assertIsInstance(elw.layout().itemAt(2).widget(), QPushButton)
        self.assertEqual(elw.layout().itemAt(2).widget(), elw.newExpButton)
        self.assertEqual(elw.newExpButton.text(), "Add new experiment")
        with patch(
            "physbiblio.gui.expWindows.ExpsListWindow.onNewExp", autospec=True
        ) as _f:
            QTest.mouseClick(elw.newExpButton, Qt.LeftButton)
            _f.assert_called_once_with(elw)
        self.assertIsInstance(elw.layout().itemAt(3).widget(), QPushButton)
        self.assertEqual(elw.layout().itemAt(3).widget(), elw.acceptButton)
        self.assertEqual(elw.acceptButton.text(), "OK")
        with patch(
            "physbiblio.gui.expWindows.ExpsListWindow.onOk", autospec=True
        ) as _f:
            QTest.mouseClick(elw.acceptButton, Qt.LeftButton)
            _f.assert_called_once_with(elw)
        self.assertIsInstance(elw.layout().itemAt(4).widget(), QPushButton)
        self.assertEqual(elw.layout().itemAt(4).widget(), elw.cancelButton)
        self.assertEqual(elw.cancelButton.text(), "Cancel")
        self.assertTrue(elw.cancelButton.autoDefault())
        with patch(
            "physbiblio.gui.expWindows.ExpsListWindow.onCancel", autospec=True
        ) as _f:
            QTest.mouseClick(elw.cancelButton, Qt.LeftButton)
            _f.assert_called_once_with(elw)

    def test_triggeredContextMenuEvent(self):
        """test triggeredContextMenuEvent"""
        p = MainWindow(testing=True)
        position = QCursor.pos()
        ev = QMouseEvent(
            QEvent.MouseButtonPress,
            position,
            Qt.RightButton,
            Qt.NoButton,
            Qt.NoModifier,
        )
        mm = PBMenu()
        mm.exec = MagicMock()
        with patch(
            "physbiblio.database.Experiments.getAll",
            return_value=self.exps,
            autospec=True,
        ) as _gh, patch(
            "physbiblio.gui.expWindows.PBMenu",
            return_value=mm,
            autospec=USE_AUTOSPEC_CLASS,
        ) as _mm:
            elw = ExpsListWindow(p)
        self.assertEqual(elw.triggeredContextMenuEvent(9999, 0, ev), None)
        self.assertEqual(elw.menu, None)
        with patch(
            "physbiblio.gui.mainWindow.MainWindow.reloadMainContent", autospec=True
        ) as _rmc, patch(
            "physbiblio.gui.expWindows.PBMenu",
            return_value=mm,
            autospec=USE_AUTOSPEC_CLASS,
        ) as _mm, patch(
            "physbiblio.gui.expWindows.editExperiment", autospec=True
        ) as _ec, patch(
            "physbiblio.gui.expWindows.deleteExperiment", autospec=True
        ) as _dc, patch(
            "physbiblio.gui.commonClasses.PBMenu.fillMenu", autospec=True
        ) as _f:
            self.assertEqual(elw.triggeredContextMenuEvent(0, 0, ev), True)
            _f.assert_called_once_with(elw.menu)
            _rmc.assert_not_called()
            _ec.assert_not_called()
            _dc.assert_not_called()

            self.assertIsInstance(elw.menu, PBMenu)
            self.assertIsInstance(elw.menu.possibleActions, list)
            self.assertEqual(len(elw.menu.possibleActions), 9)
            self.assertEqual(elw.menu.possibleActions[1], None)
            self.assertEqual(elw.menu.possibleActions[4], None)
            self.assertEqual(elw.menu.possibleActions[7], None)

            for ix, tit, en in (
                (0, "--Experiment: test0--", False),
                (2, "Open list of corresponding entries", True),
                (3, "Open in new tab", True),
                (5, "Modify", True),
                (6, "Delete", True),
                (8, "Categories", True),
            ):
                act = elw.menu.possibleActions[ix]
                self.assertIsInstance(act, QAction)
                self.assertEqual(act.text(), tit)
                self.assertEqual(act.isEnabled(), en)

            mm.exec = lambda x, i=0: mm.possibleActions[i]
            self.assertEqual(elw.triggeredContextMenuEvent(0, 0, ev), True)
            _rmc.assert_not_called()
            _ec.assert_not_called()
            _dc.assert_not_called()

            mm.exec = lambda x, i=2: mm.possibleActions[i]
            with patch(
                "physbiblio.database.Entries.getByExp",
                return_value=["a"],
                autospec=True,
            ) as _ffd:
                self.assertEqual(elw.triggeredContextMenuEvent(0, 0, ev), True)
                _ffd.assert_called_once_with(pBDB.bibs, "0")
            _rmc.assert_called_once_with(p, ["a"])
            _ec.assert_not_called()
            _dc.assert_not_called()
            _rmc.reset_mock()

            mm.exec = lambda x, i=3: mm.possibleActions[i]
            with patch(
                "physbiblio.database.Entries.getByExp",
                return_value=["a"],
                autospec=True,
            ) as _ffd:
                self.assertEqual(elw.triggeredContextMenuEvent(0, 0, ev), True)
                _ffd.assert_called_once_with(pBDB.bibs, "0")
            _rmc.assert_called_once_with(p, ["a"], newTab="test0")
            _ec.assert_not_called()
            _dc.assert_not_called()
            _rmc.reset_mock()

            mm.exec = lambda x, i=5: mm.possibleActions[i]
            with patch(
                "physbiblio.database.Entries.getByExp",
                return_value=["a"],
                autospec=True,
            ) as _ffd:
                self.assertEqual(elw.triggeredContextMenuEvent(0, 0, ev), True)
            _rmc.assert_not_called()
            _ec.assert_called_once_with(elw, p, "0")
            _dc.assert_not_called()
            _ec.reset_mock()

            mm.exec = lambda x, i=6: mm.possibleActions[i]
            with patch(
                "physbiblio.database.Entries.getByExp",
                return_value=["a"],
                autospec=True,
            ) as _ffd:
                self.assertEqual(elw.triggeredContextMenuEvent(0, 0, ev), True)
            _rmc.assert_not_called()
            _ec.assert_not_called()
            _dc.assert_called_once_with(elw, p, "0", "test0")
            _dc.reset_mock()

            with patch(
                "physbiblio.database.Experiments.getByID",
                return_value=[self.exps[0]],
                autospec=True,
            ) as _gbi, patch(
                "physbiblio.gui.catWindows.CatsTreeWindow.createForm", autospec=True
            ) as _cf:
                sc = CatsTreeWindow(
                    parent=p,
                    askCats=True,
                    askForExp=0,
                    expButton=False,
                    previous=[0, 13],
                )
            sc.exec = MagicMock()
            sc.result = "Ok"
            p.selectedCats = [9, 13]
            mm.exec = lambda x, i=8: mm.possibleActions[i]
            with patch(
                "physbiblio.database.Entries.getByExp",
                return_value=["a"],
                autospec=True,
            ) as _ffd, patch(
                "physbiblio.gui.expWindows.CatsTreeWindow",
                return_value=sc,
                autospec=USE_AUTOSPEC_CLASS,
            ) as _i, patch(
                "physbiblio.gui.mainWindow.MainWindow.statusBarMessage", autospec=True
            ) as _sbm, patch(
                "physbiblio.database.CatsExps.delete", autospec=True
            ) as _d, patch(
                "physbiblio.database.CatsExps.insert", autospec=True
            ) as _a, patch(
                "physbiblio.database.Categories.getByExp",
                return_value=[{"idCat": 0}, {"idCat": 13}],
                autospec=True,
            ) as _gbe:
                self.assertEqual(elw.triggeredContextMenuEvent(0, 0, ev), True)
                _ffd.assert_not_called()
                _gbe.assert_called_once_with(pBDB.cats, "0")
                _i.assert_called_once_with(
                    parent=p,
                    previous=[0, 13],
                    askCats=True,
                    askForExp="0",
                    expButton=False,
                )
                _d.assert_called_once_with(pBDB.catExp, 0, "0")
                _a.assert_called_once_with(pBDB.catExp, 9, "0")
                _sbm.assert_called_once_with(
                    p, "Categories for 'test0' successfully inserted"
                )
            _rmc.assert_not_called()
            _ec.assert_not_called()
            _dc.assert_not_called()

    def test_handleItemEntered(self):
        """test handleItemEntered"""
        p = QWidget()
        with patch(
            "physbiblio.database.Experiments.getAll",
            return_value=self.exps,
            autospec=True,
        ) as _gh:
            elw = ExpsListWindow(p)
        ix = elw.proxyModel.index(0, 0)
        with patch("logging.Logger.exception") as _l, patch(
            "PySide6.QtCore.QTimer.start", autospec=True
        ) as _st, patch(
            "PySide6.QtWidgets.QToolTip.showText", autospec=True
        ) as _sh, patch(
            "physbiblio.database.Experiments.getByID", return_value=[], autospec=True
        ) as _gbi:
            self.assertEqual(elw.handleItemEntered(ix), None)
            _l.assert_called_once_with("Failed in finding experiment")
            _gbi.assert_called_once_with(pBDB.exps, "0")
            _st.assert_not_called()
            _sh.assert_not_called()
        with patch("logging.Logger.exception") as _l, patch(
            "PySide6.QtCore.QTimer.start", autospec=True
        ) as _st, patch("PySide6.QtWidgets.QToolTip.showText") as _sh, patch(
            "physbiblio.database.Experiments.getByID",
            return_value=[self.exps[0]],
            autospec=True,
        ) as _gbi, patch(
            "physbiblio.database.EntryExps.countByExp", return_value=33, autospec=True
        ) as _cb, patch(
            "physbiblio.database.CatsExps.countByExp", return_value=12, autospec=True
        ) as _ce:
            position = QCursor.pos()
            self.assertEqual(elw.handleItemEntered(ix), None)
            _l.assert_not_called()
            self.assertIsInstance(elw.timer, QTimer)
            self.assertTrue(elw.timer.isSingleShot())
            _gbi.assert_called_once_with(pBDB.exps, "0")
            _st.assert_called_once_with(500)
            _sh.assert_not_called()
            elw.timer.timeout.emit()
            _sh.assert_called_once_with(
                position,
                "0: test0\nCorresponding entries: 33\nAssociated categories: 12",
                elw.tableview.viewport(),
                elw.tableview.visualRect(ix),
                3000,
            )
        with patch("logging.Logger.exception") as _l, patch(
            "PySide6.QtCore.QTimer.start", autospec=True
        ) as _st, patch("PySide6.QtWidgets.QToolTip.showText") as _sh, patch(
            "physbiblio.database.Experiments.getByID",
            return_value=[self.exps[0]],
            autospec=True,
        ) as _gbi, patch(
            "physbiblio.database.EntryExps.countByExp", return_value=33, autospec=True
        ) as _cb, patch(
            "physbiblio.database.CatsExps.countByExp", return_value=12, autospec=True
        ) as _ce:
            self.assertEqual(elw.handleItemEntered(ix), None)
            _sh.assert_called_once_with(QCursor.pos(), "", elw.tableview.viewport())

    def test_cellClick(self):
        """test cellClick"""
        p = QWidget()
        with patch(
            "physbiblio.database.Experiments.getAll",
            return_value=self.exps,
            autospec=True,
        ) as _gh:
            elw = ExpsListWindow(p)
        self.assertEqual(elw.cellClick(QModelIndex()), None)
        self.assertEqual(elw.cellClick(elw.tableModel.index(0, 0)), True)

    def test_cellDoubleClick(self):
        """test cellDoubleClick"""
        p = MainWindow(testing=True)
        exps = [
            {
                "idExp": 0,
                "name": "test0",
                "homepage": "",
                "comments": "",
                "inspire": "",
            },
            {
                "idExp": 1,
                "name": "test1",
                "homepage": "www.some.com",
                "comments": "",
                "inspire": "1234",
            },
        ]
        with patch(
            "physbiblio.database.Experiments.getAll", return_value=exps, autospec=True
        ) as _gh:
            elw = ExpsListWindow(p)
        self.assertEqual(elw.cellDoubleClick(QModelIndex()), None)
        with patch(
            "physbiblio.gui.mainWindow.MainWindow.reloadMainContent", autospec=True
        ) as _rmc, patch(
            "physbiblio.database.Entries.getByExp", return_value=["a"], autospec=True
        ) as _ffd:
            self.assertEqual(elw.cellDoubleClick(elw.proxyModel.index(0, 0)), True)
            self.assertEqual(elw.cellDoubleClick(elw.proxyModel.index(0, 1)), True)
            self.assertEqual(elw.cellDoubleClick(elw.proxyModel.index(0, 2)), True)
            self.assertEqual(elw.cellDoubleClick(elw.proxyModel.index(0, 3)), None)
            self.assertEqual(elw.cellDoubleClick(elw.proxyModel.index(0, 4)), None)
            self.assertEqual(elw.cellDoubleClick(elw.proxyModel.index(1, 0)), True)
            self.assertEqual(elw.cellDoubleClick(elw.proxyModel.index(1, 1)), True)
            self.assertEqual(elw.cellDoubleClick(elw.proxyModel.index(1, 2)), True)
            self.assertEqual(_rmc.call_count, 8)
            _rmc.assert_has_calls([call(p, ["a"])])
        with patch(
            "physbiblio.gui.commonClasses.GUIViewEntry.openLink", autospec=True
        ) as _ol, patch("logging.Logger.debug") as _l:
            self.assertEqual(elw.cellDoubleClick(elw.proxyModel.index(1, 3)), True)
            _ol.assert_called_once_with(pBGuiView, "www.some.com", "link")
            _l.assert_called_once_with("Opening 'www.some.com'...")
        with patch(
            "physbiblio.gui.commonClasses.GUIViewEntry.openLink", autospec=True
        ) as _ol, patch("logging.Logger.debug") as _l:
            self.assertEqual(elw.cellDoubleClick(elw.proxyModel.index(1, 4)), True)
            _ol.assert_called_once_with(
                pBGuiView, pbConfig.inspireExperimentsLink + "1234", "link"
            )
            _l.assert_called_once_with(
                "Opening '%s'..." % (pbConfig.inspireExperimentsLink + "1234")
            )
        with patch(
            "physbiblio.gui.commonClasses.GUIViewEntry.openLink",
            side_effect=Exception("some error"),
            autospec=True,
        ) as _ol, patch("logging.Logger.debug") as _l, patch(
            "logging.Logger.warning"
        ) as _w:
            self.assertEqual(elw.cellDoubleClick(elw.proxyModel.index(1, 4)), True)
            _ol.assert_called_once_with(
                pBGuiView, pbConfig.inspireExperimentsLink + "1234", "link"
            )
            _l.assert_called_once_with(
                "Opening '%s'..." % (pbConfig.inspireExperimentsLink + "1234")
            )
            _w.assert_called_once_with(
                "Opening link '%s' failed!"
                % (pbConfig.inspireExperimentsLink + "1234"),
                exc_info=True,
            )


@unittest.skipIf(skipTestsSettings.gui, "GUI tests")
class TestEditExperimentDialog(GUITestCase):
    """Test EditExperimentDialog"""

    def test_init(self):
        """test init"""
        p = QWidget()
        with patch(
            "physbiblio.gui.expWindows.EditExperimentDialog.createForm", autospec=True
        ) as _cf:
            eed = EditExperimentDialog(p)
            _cf.assert_called_once_with(eed)
        self.assertIsInstance(eed, EditObjectWindow)
        self.assertEqual(eed.parent(), p)
        self.assertIsInstance(eed.data, dict)
        self.assertEqual(
            eed.data,
            {"idExp": "", "name": "", "homepage": "", "inspire": "", "comments": ""},
        )
        exp = {
            "idExp": "9999",
            "name": "myexp",
            "homepage": "www.some.exp",
            "inspire": "1234",
            "comments": "no comments",
        }
        with patch(
            "physbiblio.gui.expWindows.EditExperimentDialog.createForm", autospec=True
        ) as _cf:
            eed = EditExperimentDialog(parent=p, experiment=exp)
            _cf.assert_called_once_with(eed)
        self.assertEqual(eed.data, exp)

    def test_createForm(self):
        """test createForm"""
        p = QWidget()
        eed = EditExperimentDialog(p)
        self.assertEqual(eed.windowTitle(), "Edit experiment")

        self.assertIsInstance(eed.layout().itemAtPosition(1, 0).widget(), PBLabel)
        self.assertEqual(eed.layout().itemAtPosition(1, 0).widget().text(), "name")
        self.assertIsInstance(eed.layout().itemAtPosition(1, 1).widget(), PBLabel)
        self.assertEqual(
            eed.layout().itemAtPosition(1, 1).widget().text(),
            "(Name of the experiment)",
        )
        self.assertIsInstance(eed.layout().itemAtPosition(2, 0).widget(), QLineEdit)
        self.assertEqual(
            eed.layout().itemAtPosition(2, 0).widget(), eed.textValues["name"]
        )
        self.assertEqual(eed.textValues["name"].text(), "")
        self.assertEqual(eed.textValues["name"].isEnabled(), True)

        self.assertIsInstance(eed.layout().itemAtPosition(3, 0).widget(), PBLabel)
        self.assertEqual(eed.layout().itemAtPosition(3, 0).widget().text(), "comments")
        self.assertIsInstance(eed.layout().itemAtPosition(3, 1).widget(), PBLabel)
        self.assertEqual(
            eed.layout().itemAtPosition(3, 1).widget().text(),
            "(Description or comments)",
        )
        self.assertIsInstance(eed.layout().itemAtPosition(4, 0).widget(), QLineEdit)
        self.assertEqual(
            eed.layout().itemAtPosition(4, 0).widget(), eed.textValues["comments"]
        )
        self.assertEqual(eed.textValues["comments"].text(), "")
        self.assertEqual(eed.textValues["comments"].isEnabled(), True)

        self.assertIsInstance(eed.layout().itemAtPosition(5, 0).widget(), PBLabel)
        self.assertEqual(eed.layout().itemAtPosition(5, 0).widget().text(), "homepage")
        self.assertIsInstance(eed.layout().itemAtPosition(5, 1).widget(), PBLabel)
        self.assertEqual(
            eed.layout().itemAtPosition(5, 1).widget().text(),
            "(Web link to the experiment homepage)",
        )
        self.assertIsInstance(eed.layout().itemAtPosition(6, 0).widget(), QLineEdit)
        self.assertEqual(
            eed.layout().itemAtPosition(6, 0).widget(), eed.textValues["homepage"]
        )
        self.assertEqual(eed.textValues["homepage"].text(), "")
        self.assertEqual(eed.textValues["homepage"].isEnabled(), True)

        self.assertIsInstance(eed.layout().itemAtPosition(7, 0).widget(), PBLabel)
        self.assertEqual(eed.layout().itemAtPosition(7, 0).widget().text(), "inspire")
        self.assertIsInstance(eed.layout().itemAtPosition(7, 1).widget(), PBLabel)
        self.assertEqual(
            eed.layout().itemAtPosition(7, 1).widget().text(),
            "(INSPIRE-HEP ID of the experiment record)",
        )
        self.assertIsInstance(eed.layout().itemAtPosition(8, 0).widget(), QLineEdit)
        self.assertEqual(
            eed.layout().itemAtPosition(8, 0).widget(), eed.textValues["inspire"]
        )
        self.assertEqual(eed.textValues["inspire"].text(), "")
        self.assertEqual(eed.textValues["inspire"].isEnabled(), True)

        self.assertIsInstance(eed.layout().itemAtPosition(9, 0).widget(), QPushButton)
        self.assertEqual(eed.layout().itemAtPosition(9, 0).widget(), eed.acceptButton)
        self.assertEqual(eed.acceptButton.text(), "OK")
        self.assertIsInstance(eed.layout().itemAtPosition(9, 1).widget(), QPushButton)
        self.assertEqual(eed.layout().itemAtPosition(9, 1).widget(), eed.cancelButton)
        self.assertEqual(eed.cancelButton.text(), "Cancel")
        self.assertTrue(eed.cancelButton.autoDefault())
        with patch(
            "physbiblio.gui.commonClasses.EditObjectWindow.onOk", autospec=True
        ) as _f:
            QTest.mouseClick(eed.acceptButton, Qt.LeftButton)
            _f.assert_called_once_with(eed)
        with patch(
            "physbiblio.gui.commonClasses.EditObjectWindow.onCancel", autospec=True
        ) as _f:
            QTest.mouseClick(eed.cancelButton, Qt.LeftButton)
            _f.assert_called_once_with(eed)

        exp = {
            "idExp": "9999",
            "name": "myexp",
            "homepage": "www.some.exp",
            "inspire": "1234",
            "comments": "no comments",
        }
        eed = EditExperimentDialog(parent=p, experiment=exp)

        self.assertIsInstance(eed.layout().itemAtPosition(1, 0).widget(), PBLabel)
        self.assertEqual(eed.layout().itemAtPosition(1, 0).widget().text(), "idExp")
        self.assertIsInstance(eed.layout().itemAtPosition(1, 1).widget(), PBLabel)
        self.assertEqual(
            eed.layout().itemAtPosition(1, 1).widget().text(),
            "(Unique ID that identifies the experiment)",
        )
        self.assertIsInstance(eed.layout().itemAtPosition(2, 0).widget(), QLineEdit)
        self.assertEqual(
            eed.layout().itemAtPosition(2, 0).widget(), eed.textValues["idExp"]
        )
        self.assertEqual(eed.textValues["idExp"].text(), "9999")
        self.assertEqual(eed.textValues["idExp"].isEnabled(), False)

        self.assertIsInstance(eed.layout().itemAtPosition(3, 0).widget(), PBLabel)
        self.assertEqual(eed.layout().itemAtPosition(3, 0).widget().text(), "name")
        self.assertIsInstance(eed.layout().itemAtPosition(3, 1).widget(), PBLabel)
        self.assertEqual(
            eed.layout().itemAtPosition(3, 1).widget().text(),
            "(Name of the experiment)",
        )
        self.assertIsInstance(eed.layout().itemAtPosition(4, 0).widget(), QLineEdit)
        self.assertEqual(
            eed.layout().itemAtPosition(4, 0).widget(), eed.textValues["name"]
        )
        self.assertEqual(eed.textValues["name"].text(), "myexp")
        self.assertEqual(eed.textValues["name"].isEnabled(), True)

        self.assertIsInstance(eed.layout().itemAtPosition(5, 0).widget(), PBLabel)
        self.assertEqual(eed.layout().itemAtPosition(5, 0).widget().text(), "comments")
        self.assertIsInstance(eed.layout().itemAtPosition(5, 1).widget(), PBLabel)
        self.assertEqual(
            eed.layout().itemAtPosition(5, 1).widget().text(),
            "(Description or comments)",
        )
        self.assertIsInstance(eed.layout().itemAtPosition(6, 0).widget(), QLineEdit)
        self.assertEqual(
            eed.layout().itemAtPosition(6, 0).widget(), eed.textValues["comments"]
        )
        self.assertEqual(eed.textValues["comments"].text(), "no comments")
        self.assertEqual(eed.textValues["comments"].isEnabled(), True)

        self.assertIsInstance(eed.layout().itemAtPosition(7, 0).widget(), PBLabel)
        self.assertEqual(eed.layout().itemAtPosition(7, 0).widget().text(), "homepage")
        self.assertIsInstance(eed.layout().itemAtPosition(7, 1).widget(), PBLabel)
        self.assertEqual(
            eed.layout().itemAtPosition(7, 1).widget().text(),
            "(Web link to the experiment homepage)",
        )
        self.assertIsInstance(eed.layout().itemAtPosition(8, 0).widget(), QLineEdit)
        self.assertEqual(
            eed.layout().itemAtPosition(8, 0).widget(), eed.textValues["homepage"]
        )
        self.assertEqual(eed.textValues["homepage"].text(), "www.some.exp")
        self.assertEqual(eed.textValues["homepage"].isEnabled(), True)

        self.assertIsInstance(eed.layout().itemAtPosition(9, 0).widget(), PBLabel)
        self.assertEqual(eed.layout().itemAtPosition(9, 0).widget().text(), "inspire")
        self.assertIsInstance(eed.layout().itemAtPosition(9, 1).widget(), PBLabel)
        self.assertEqual(
            eed.layout().itemAtPosition(9, 1).widget().text(),
            "(INSPIRE-HEP ID of the experiment record)",
        )
        self.assertIsInstance(eed.layout().itemAtPosition(10, 0).widget(), QLineEdit)
        self.assertEqual(
            eed.layout().itemAtPosition(10, 0).widget(), eed.textValues["inspire"]
        )
        self.assertEqual(eed.textValues["inspire"].text(), "1234")
        self.assertEqual(eed.textValues["inspire"].isEnabled(), True)
        self.assertEqual(eed.layout().itemAtPosition(11, 0).widget(), eed.acceptButton)
        self.assertEqual(eed.layout().itemAtPosition(11, 1).widget(), eed.cancelButton)


if __name__ == "__main__":
    unittest.main()



"""Module for retrieving bibtex information
from a list of given web resources
(see `__all__`).

This module is part of the physbiblio package.
"""
from physbiblio import __author__, __email__, __version__, __version_date__

__all__ = [
    "adsnasa",
    "arxiv",
    "doi",
    "inspire",
    "webInterf",
    "tests",
]



"""Module that deals with importing info from the INSPIRE-HEP API.

This file is part of the physbiblio package.
"""
import json
import re
import time
import traceback

import bibtexparser

try:
    from physbiblio.bibtexWriter import pbWriter
    from physbiblio.config import pbConfig
    from physbiblio.errors import pBLogger
    from physbiblio.parseAccents import parse_accents_str
    from physbiblio.strings.webimport import InspireStrings
    from physbiblio.webimport.arxiv import getYear
    from physbiblio.webimport.webInterf import WebInterf, physBiblioWeb
except ImportError:
    print("Could not find physbiblio and its modules!")
    print(traceback.format_exc())
    raise


class WebSearch(WebInterf, InspireStrings):
    """Subclass of WebInterf that can connect
    to INSPIRE-HEP to perform searches
    """

    name = "inspire"
    description = "INSPIRE fetcher"
    url = pbConfig.inspireLiteratureAPI
    urlRecord = pbConfig.inspireLiteratureLink
    correspondences = [
        ["id", "inspire"],
        ["year", "year"],
        # ["arxiv", "arxiv"],
        # ["oldkeys", "old_keys"],
        ["firstdate", "firstdate"],
        ["pubdate", "pubdate"],
        ["doi", "doi"],
        ["ads", "ads"],
        ["isbn", "isbn"],
        ["bibtex", "bibtex"],
        ["link", "link"],
        ["cit", "citations"],
        ["cit_no_self", "citations_no_self"],
    ]
    bibtexFields = [
        "author",
        "title",
        "journal",
        "volume",
        "year",
        "pages",
        # "arxiv",
        "primaryclass",
        "archiveprefix",
        "eprint",
        "doi",
        "isbn",
        "school",
        "reportnumber",
        "booktitle",
        "collaboration",
    ]
    updateBibtexFields = [
        "author",
        "title",
        "doi",
        "volume",
        "pages",
        "year",
        "journal",
        "eprint",
        "archiveprefix",
        "primaryclass",
        "collaboration",
        "reportnumber",
        "booktitle",
        "publisher",
        "arxiv",
    ]
    metadataLiteratureFields = [
        "arxiv_eprints",
        "author_count",
        "authors.full_name",
        "citation_count",
        "citation_count_without_self_citations",
        "collaborations",
        "control_number",
        "document_type",
        "dois",
        "earliest_date",
        "external_system_identifiers",
        "first_author.full_name",
        "imprints",
        "isbns",
        "legacy_creation_date",
        "preprint_date",
        "primary_arxiv_category",
        "publication_info",
        "report_numbers",
        "texkeys",
        "thesis_info",
        "titles",
    ]
    metadataCitationFields = [
        "citation_count",
        "citation_count_without_self_citations",
        "control_number",
    ]
    metadataConferenceFields = [
        "cnum",
        "proceedings",
        "titles",
    ]
    defaultSize = 250

    def __init__(self):
        """Initializes the class variables
        using the WebInterf constructor.

        Define additional specific parameters for the INSPIRE-HEP API.
        """
        WebInterf.__init__(self)
        self.urlArgs = {
            "sort": "mostrecent",
            "size": "%s" % self.defaultSize,
            "page": "1",
        }

    def _fixAccents(self, text):
        """Enclose particular accent strings such as '\"a' in curly braces
        to avoid parsing errors in bibtexparser

        Parameter:
            text: the input bibtex

        Output:
            the fixed bibtex
        """
        match = re.compile('((\\\\")[a-zA-Z]{1})', re.MULTILINE)
        for t in match.finditer(text):
            text = text.replace(t.group(), "{%s}" % t.group())
        return text

    def retrieveBibtex(self, string, size=250):
        """Retrieves a list of bibtexs

        Parameters:
            string: the search string

        Output:
            returns the bibtex string obtained from the API
        """
        args = self.urlArgs.copy()
        args["q"] = string.replace(" ", "%20")
        args["size"] = str(size)
        args["format"] = "bibtex"
        url = self.createUrl(args)
        pBLogger.info(self.searchInfo % (string, url))
        text = self.textFromUrl(url)
        text = self._fixAccents(text)
        try:
            return parse_accents_str(text)
        except Exception:
            pBLogger.exception(self.genericError)
            return ""

    def retrieveUrlFirst(self, string):
        """Retrieves the first result
        from the content of the given web page.

        Parameters:
            string: the search string

        Output:
            returns the bibtex string obtained from the API
        """
        return self.retrieveBibtex(string, size=1)

    def retrieveUrlAll(self, string):
        """Retrieves all the result
        from the content of the given web page.

        Parameters:
            string: the search string

        Output:
            returns the bibtex string obtained from the API
        """
        return self.retrieveBibtex(string, size=self.defaultSize)

    def retrieveAPIResults(self, url, max_iterations=20):
        """Obtain results from the INSPIRE API, merging several calls
        if more than one page is available

        Parameter:
            url: the initial url to fetch.
                Subsequent ones, if present, will be fetched by the json
            max_iterations (default 20): maximum number of times
                that links.next is employed before stopping

        Output:
            a list of hits and the total number of hits
                as returned by the first GET
        """
        hits = []
        tot = 0
        iteration = 0
        while url != "" and url is not None and iteration < max_iterations:
            pBLogger.info(self.searchResultsFrom % (url))
            text = self.textFromUrl(url)
            if text is None or text == "":
                pBLogger.warning(self.errorEmptyText)
                return hits, tot
            try:
                results = json.loads(text)
            except json.decoder.JSONDecodeError:
                pBLogger.exception(self.jsonError)
                return hits, tot
            if "message" in results.keys():
                pBLogger.exception(
                    self.apiResponseError % (results["status"], results["message"])
                )
                return hits, tot
            elif "id" in results.keys():
                return [results], 1
            elif "hits" in results.keys():
                try:
                    hits += results["hits"]["hits"]
                    if tot == 0:
                        tot = results["hits"]["total"]
                except (IndexError, KeyError):
                    pBLogger.exception(self.genericError)
                    return hits, tot
                try:
                    url = results["links"]["next"]
                    iteration += 1
                except KeyError:
                    url = ""
        return hits, tot

    def retrieveSearchResults(
        self, searchstring, size=500, fields=None, addfields=[], max_iterations=20
    ):
        """Extract a list of hits from an Inspire search
        (through the q parameter)

        Parameters:
            searchstring: the string used to query the Inspire API
            size (default 500): how many hits per page (max 1000)
            fields (default None): if not None, replace the default selection
                of metadata fields required in the output
            addfields (default []): list of requested metadata fields
                apart from the default ones
                (useful especially when fields=None)
            max_iterations (default 20): maximum number of times
                that links.next is employed before stopping

        Output:
            from self.retrieveAPIResults
        """
        if fields is None:
            fields = self.metadataLiteratureFields
        if len(addfields) > 0:
            fields += addfields
        args = self.urlArgs.copy()
        args["q"] = searchstring.replace(" ", "%20")
        args["size"] = "%d" % (size if size <= 1000 else 1000)
        args["fields"] = ",".join(fields)
        try:
            del args["page"]
        except KeyError:
            pass
        return self.retrieveAPIResults(
            self.createUrl(args), max_iterations=max_iterations
        )

    def retrieveBatchQuery(self, entries, searchFormat="%s", **kwargs):
        """Read a list of entries and create a general search string
        for a "manual" batch query to the INSPIRE database
        through the new API.

        Parameters:
            entries: a list search strings for the entries of interest
            searchFormat (default "%s"): formatter for the search string
                of each entry. Example: use "recid:%s" to match IDs
            kwargs: passed to self.retrieveSearchResults

        Output:
            the output of self.retrieveSearchResults
        """
        entries = [e for e in entries if e != "" and e is not None]
        try:
            if searchFormat == "%s":
                entries = [
                    "texkeys:%s" % e if e.lower().startswith("de:") else e
                    for e in entries
                ]
        except (AttributeError, TypeError):
            pass
        searchString = " or ".join([searchFormat % e for e in entries])
        return self.retrieveSearchResults(searchString, **kwargs)

    def retrieveInspireID(self, string, number=None, isDoi=False, isArxiv=False):
        """Read the fetched content for a given entry
        to obtain its INSPIRE-HEP ID

        Parameters:
            string: the search string
            number (optional): the integer corresponding
                to the desired entry in the list,
                if more than one is present
            isDoi (default False): if the search string is a DOI,
                add the "doi" prefix in the url
            isArxiv (default False): if the search string is an arxiv id,
                add the "eprint" (and "arxiv:" if needed) prefix in the url
        """
        i = 0
        if isDoi:
            string = "doi+" + string
        elif isArxiv:
            if re.match("[0-9]{4}\.[0-9]{4,5}", string) and "arxiv:" not in string:
                string = "arxiv:" + string
            string = "eprint " + string
        args = self.urlArgs.copy()
        args["q"] = string.replace(" ", "%20")
        args["size"] = "1"
        args["fields"] = "control_number"
        try:
            args["page"] = int(number)
            if args["page"] == 0:
                del args["page"]
        except (TypeError, ValueError):
            pass
        url = self.createUrl(args)
        pBLogger.info(self.searchIDInfo % (string, url))
        text = self.textFromUrl(url)
        if text is None or text.strip() == "":
            pBLogger.warning(self.errorEmptyText)
            return ""
        try:
            results = json.loads(text)
        except json.decoder.JSONDecodeError:
            pBLogger.exception(self.jsonError)
            return ""
        try:
            inspireID = "%s" % results["hits"]["hits"][0]["metadata"]["control_number"]
        except (IndexError, KeyError):
            pBLogger.exception(self.genericError)
            return ""
        pBLogger.info(self.foundID % inspireID)
        return inspireID

    def retrieveCumulativeUpdates(self, date1, date2, max_iterations=20):
        """Harvest the INSPRIE API to get all the updates
        and new occurrences between two dates

        Parameters:
            date1, date2: dates that define
                the time interval to be searched
            max_iterations (default 20): maximum number of times
                that links.next is employed before stopping

        Output:
            a list of dictionaries containing the bibtex information
        """
        pBLogger.info(self.startString % time.strftime("%c"))
        hits, total = self.retrieveSearchResults(
            "du%3E%3D" + date1 + " and du%3C%3D" + date2, max_iterations=max_iterations
        )
        foundObjects = []
        if len(hits) != total:
            pBLogger.info(self.numberOfEntriesChanged % (len(hits), total))
        for count, rec in enumerate(hits):
            id = rec["metadata"]["control_number"]
            if count % 500 == 0:
                pBLogger.info(self.processed % count)
            if not rec:
                continue
            try:
                tmpDict = self.readRecord(rec)
                tmpDict["id"] = "%s" % id
                foundObjects.append(tmpDict)
            except Exception as e:
                pBLogger.exception(self.exceptionFormat % (count, id, e))
        pBLogger.info(self.processed % len(hits))
        pBLogger.info(self.endString % time.strftime("%c"))
        return foundObjects

    def processRecord(self, record, bibtex=None, readConferenceTitle=False):
        """Process the JSON entry for a given record

        Parameters:
            record: the INSPIRE-HEP JSON record
            bibtex (default None): whether the bibtex should be
                included in the output dictionary
            readConferenceTitle (boolean, default False):
                try to read the conference title if dealing
                with a proceeding

        Output:
            the dictionary containing the bibtex information
        """
        try:
            res = self.readRecord(record, readConferenceTitle=readConferenceTitle)
        except Exception:
            pBLogger.exception(self.errorReadRecord)
            return False
        if bibtex is not None:
            try:
                outcome, bibtex = self.updateBibtex(res, bibtex)
            except Exception:
                pBLogger.exception(self.errorUpdateBibtex)
                return False
            if outcome:
                res["bibtex"] = bibtex
        return res

    def retrieveOAIData(
        self, inspireID, bibtex=None, verbose=0, readConferenceTitle=False
    ):
        """Get the JSON entry for a given record

        Parameters:
            inspireID: the INSPIRE-HEP identifier (a number)
                of the desired entry
            bibtex (default None): whether the bibtex should be
                included in the output dictionary
            verbose (default 0): increase the output level
            readConferenceTitle (boolean, default False):
                try to read the conference title if dealing
                with a proceeding

        Output:
            the dictionary containing the bibtex information
        """
        hits, tot = self.retrieveAPIResults("%s%s" % (self.url, inspireID))
        if verbose > 0:
            pBLogger.info(self.readData + time.strftime("%c"))
        try:
            record = hits[0]
        except IndexError:
            pBLogger.exception(self.errorEmptySearch)
            return False
        try:
            res = self.processRecord(
                record, bibtex=bibtex, readConferenceTitle=readConferenceTitle
            )
        except Exception:
            pBLogger.exception(self.errorReadRecord)
            return False
        if verbose > 0:
            pBLogger.info(self.doneD)
        return res

    def getProceedingsTitle(self, conferenceCode, useUrl=None):
        """Use INSPIRE-HEP API to retrieve the title
        of the Proceedings associated to a conference
        identified by `conferenceCode`

        Parameters:
            conferenceCode: the identifier of the conference
                in the INSPIRE-HEP database
            useUrl (default None): if not None, the link to use
                for retrieving the title of conference proceedings

        Output:
            a string, if found, or None
        """
        if useUrl is None:
            url = (
                pbConfig.inspireConferencesAPI
                + "?q=%s" % conferenceCode
                + "&fields="
                + (",".join(self.metadataConferenceFields))
            )
            text = self.textFromUrl(url)
            try:
                info = json.loads(text)
            except (TypeError, json.decoder.JSONDecodeError):
                pBLogger.exception(self.jsonError)
                return None
            try:
                confs = [
                    a
                    for a in info["hits"]["hits"]
                    if a["metadata"]["cnum"] == conferenceCode
                ]
            except (KeyError, TypeError):
                return None
            try:
                procid = confs[0]["metadata"]["proceedings"][0]["control_number"]
            except (IndexError, KeyError, TypeError):
                return None
            url = "%s%s" % (pbConfig.inspireLiteratureAPI, procid)
        else:
            url = useUrl
        text = self.textFromUrl(url)
        try:
            info = json.loads(text)
        except json.decoder.JSONDecodeError:
            pBLogger.exception(self.jsonError)
            return None
        try:
            title = "%s: %s" % (
                info["metadata"]["titles"][0]["title"],
                info["metadata"]["titles"][0]["subtitle"],
            )
        except (IndexError, KeyError, TypeError):
            return None
        return title

    def readRecord(self, record, readConferenceTitle=False, noWarning=False):
        """Read the content of a marcxml record
        to return a bibtex string

        Parameters:
            record: the marcxml record to read
            readConferenceTitle (default False): if True, look for
                the proceedings info to get the title of the conference
            noWarning (default False): if True, suppress some warnings

        Output:
            a dictionary with the obtained fields
        """

        tmpDict = {}
        try:
            tmpDict["id"] = record["metadata"]["control_number"]
        except (KeyError, TypeError):
            tmpDict["id"] = None
        # key
        tmpDict["bibkey"] = None
        try:
            tmpDict["bibkey"] = record["metadata"]["texkeys"][0]
        except (IndexError, KeyError, TypeError):
            if not noWarning:
                pBLogger.warning(self.errorReadRecord % tmpDict["id"])
        # old keys
        tmpOld = []
        try:
            tmpOld = list(record["metadata"]["texkeys"])
            tmpOld.pop(0)
        except (IndexError, KeyError, TypeError):
            if not noWarning:
                pBLogger.warning(self.errorReadRecord % tmpDict["id"])
        tmpDict["oldkeys"] = ",".join(tmpOld)
        # doi
        tmpDict["doi"] = None
        try:
            tmpDict["doi"] = record["metadata"]["dois"][0]["value"]
        except (IndexError, KeyError, TypeError):
            pass
        # arxiv
        tmpDict["eprint"] = None
        tmpDict["archiveprefix"] = None
        try:
            tmpDict["eprint"] = record["metadata"]["arxiv_eprints"][0]["value"]
            tmpDict["archiveprefix"] = "arXiv"
        except (IndexError, KeyError, TypeError):
            pass
        tmpDict["primaryclass"] = None
        try:
            tmpDict["primaryclass"] = record["metadata"]["primary_arxiv_category"][0]
        except (IndexError, KeyError):
            try:
                tmpDict["primaryclass"] = record["metadata"]["arxiv_eprints"][0][
                    "categories"
                ][0]
            except (IndexError, KeyError, TypeError):
                pass
        # ads
        tmpDict["ads"] = None
        try:
            tmpDict["ads"] = [
                a["value"]
                for a in record["metadata"]["external_system_identifiers"]
                if a["schema"] == "ADS"
            ][0]
        except (IndexError, KeyError, TypeError):
            pass
        # authors
        try:
            tmpDict["author"] = record["metadata"]["first_author"]["full_name"]
        except KeyError:
            try:
                tmpDict["author"] = record["metadata"]["authors"][0]["full_name"]
            except (IndexError, KeyError, TypeError):
                tmpDict["author"] = ""
        try:
            if (
                record["metadata"]["author_count"]
                if "author_count" in record["metadata"].keys()
                else len(record["metadata"]["authors"])
            ) > pbConfig.params["maxAuthorSave"] - 1:
                tmpDict["author"] += " and others"
            else:
                for r in record["metadata"]["authors"][1:]:
                    tmpDict["author"] += " and %s" % r["full_name"]
        except (IndexError, KeyError, TypeError):
            pass
        # collaboration
        try:
            tmpDict["collaboration"] = ", ".join(
                [c["value"] for c in record["metadata"]["collaborations"]]
            )
        except (KeyError, TypeError):
            tmpDict["collaboration"] = None
        # title
        try:
            tmpDict["title"] = record["metadata"]["titles"][0]["title"]
        except (IndexError, KeyError, TypeError):
            tmpDict["title"] = None
        # conference title
        try:
            pi = record["metadata"]["publication_info"][0]
            conferenceCode = pi["cnum"] if "cnum" in pi.keys() else None
        except (IndexError, KeyError, TypeError):
            conferenceCode = None
        try:
            pi = record["metadata"]["publication_info"][0]
            parentUrl = pi["parent_record"]["$ref"]
        except (IndexError, KeyError, TypeError):
            parentUrl = None
        if conferenceCode is not None and readConferenceTitle:
            tmpDict["booktitle"] = self.getProceedingsTitle(
                conferenceCode, useUrl=parentUrl
            )
        # publication info
        tmpDict["year"] = None
        if tmpDict["eprint"] is not None and tmpDict["year"] is None:
            tmpDict["year"] = getYear(tmpDict["eprint"])
        try:
            pi = record["metadata"]["publication_info"][0]
            try:
                tmpDict["journal"] = pi["journal_title"]
            except KeyError:
                tmpDict["journal"] = None
            try:
                tmpDict["volume"] = pi["journal_volume"]
            except KeyError:
                tmpDict["volume"] = None
            try:
                tmpDict["year"] = pi["year"]
            except KeyError:
                pass
            try:
                tmpDict["pages"] = (
                    pi["artid"]
                    if "artid" in pi.keys()
                    else "%s-%s" % (pi["page_start"], pi["page_end"])
                    if ("page_start" in pi.keys() and "page_end" in pi.keys())
                    else pi["page_start"]
                    if "page_start" in pi.keys()
                    else None
                )
            except KeyError:
                tmpDict["pages"] = None
        except (IndexError, KeyError, TypeError):
            tmpDict["journal"] = None
            tmpDict["volume"] = None
            tmpDict["pages"] = None
        # dates
        try:
            tmpDict["firstdate"] = (
                record["metadata"]["preprint_date"]
                if "preprint_date" in record["metadata"].keys()
                else record["metadata"]["legacy_creation_date"]
                if "legacy_creation_date" in record["metadata"].keys()
                else record["metadata"]["earliest_date"]
            )
        except (KeyError, TypeError):
            tmpDict["firstdate"] = None
        try:
            tmpDict["pubdate"] = record["metadata"]["imprints"][0]["date"]
        except (IndexError, KeyError, TypeError):
            tmpDict["pubdate"] = None
        # report numbers
        tmpDict["reportnumber"] = None
        try:
            tmpDict["reportnumber"] = ", ".join(
                [c["value"] for c in record["metadata"]["report_numbers"]]
            )
        except (KeyError, TypeError):
            tmpDict["reportnumber"] = None
        # isbn
        try:
            tmpDict["isbn"] = record["metadata"]["isbns"][0]["value"]
        except (IndexError, KeyError, TypeError):
            tmpDict["isbn"] = None
        # entry type
        if tmpDict["isbn"] is not None:
            tmpDict["ENTRYTYPE"] = "book"
        else:
            try:
                collections = record["metadata"]["document_type"]
            except KeyError:
                collections = []
            if "conference paper" in collections or conferenceCode is not None:
                tmpDict["ENTRYTYPE"] = "inproceedings"
            elif "thesis" in collections:
                tmpDict["ENTRYTYPE"] = "phdthesis"
                try:
                    tmpDict["school"] = ", ".join(
                        record["metadata"]["thesis_info"]["institutions"]
                    )
                    tmpDict["year"] = record["metadata"]["thesis_info"]["date"]
                except (KeyError, TypeError):
                    pass
            else:
                tmpDict["ENTRYTYPE"] = "article"
        # clean accents
        for k in tmpDict.keys():
            try:
                tmpDict[k] = parse_accents_str(tmpDict[k])
            except Exception:
                pass
        # citations
        try:
            tmpDict["cit_no_self"] = int(
                record["metadata"]["citation_count_without_self_citations"]
            )
        except (KeyError, ValueError):
            tmpDict["cit_no_self"] = None
        try:
            tmpDict["cit"] = int(record["metadata"]["citation_count"])
        except (KeyError, ValueError):
            tmpDict["cit"] = None
        # link
        tmpDict["link"] = None
        try:
            if tmpDict["eprint"] is not None and tmpDict["eprint"] != "":
                tmpDict["link"] = pbConfig.arxivUrl + "/abs/" + tmpDict["eprint"]
        except KeyError:
            pass
        try:
            if tmpDict["doi"] is not None and tmpDict["doi"] != "":
                tmpDict["link"] = pbConfig.doiUrl + tmpDict["doi"]
        except KeyError:
            pass
        # build bibtex
        bibtexDict = {"ENTRYTYPE": tmpDict["ENTRYTYPE"], "ID": tmpDict["bibkey"]}
        for k in self.bibtexFields:
            if k in tmpDict.keys() and tmpDict[k] is not None and tmpDict[k] != "":
                bibtexDict[k] = tmpDict[k]
        db = bibtexparser.bibdatabase.BibDatabase()
        db.entries = [bibtexDict]
        tmpDict["bibtex"] = pbWriter.write(db)
        return tmpDict

    def updateBibtex(self, res, bibtex, force=False):
        """use OAI data to update the (existing) bibtex information.
        Unless force is True,
        update record only if publication info is present
        (assume data won't change unless the paper is published)

        Parameters:
            res: the recent search results
            bibtex: the old bibtex to be updated

        Output:
            True/False and a string containing the bibtex of the entry
        """

        def arxivEprint(element):
            """eprint has priority over arxiv field:
            if the latter is present, delete it,
            eventually copying its content into the former
            """
            if "arxiv" in element.keys():
                if "eprint" not in element.keys():
                    element["eprint"] = element["arxiv"]
                del element["arxiv"]
            return element

        try:
            element = bibtexparser.loads(bibtex).entries[0]
        except Exception:
            pBLogger.warning(self.errorInvalidBibtex % bibtex)
            return False, bibtex
        if not force:
            try:
                assert res["journal"] is not None
            except (AssertionError, KeyError):
                pBLogger.warning(self.warningJournal % (res["id"]))
                return False, bibtex
        try:
            newelement = bibtexparser.loads(res["bibtex"]).entries[0]
        except KeyError:
            newelement = {}
        except Exception:
            pBLogger.warning(self.errorInvalidBibtex % res["bibtex"])
            newelement = {}
        element = {**element, **newelement}
        element = arxivEprint(element)
        if not all([k in res for k in self.updateBibtexFields]):
            pBLogger.warning(
                self.warningMissingField
                % (
                    [k for k in self.updateBibtexFields if k not in res.keys()],
                    res["id"],
                )
            )
        for k in self.updateBibtexFields:
            if k in res and res[k] != "" and res[k] is not None:
                element[k] = res[k]
        element = arxivEprint(element)
        db = bibtexparser.bibdatabase.BibDatabase()
        db.entries = [element]
        return True, pbWriter.write(db)



"""Module that deals with importing info from the arXiv API.

Uses feedparser module to read the page content.

This file is part of the physbiblio package.
"""
import re
import traceback

import feedparser

try:
    from bibtexparser.bibdatabase import BibDatabase

    from physbiblio.bibtexWriter import pbWriter
    from physbiblio.config import pbConfig
    from physbiblio.errors import pBLogger
    from physbiblio.parseAccents import parse_accents_str
    from physbiblio.strings.webimport import ArxivStrings
    from physbiblio.webimport.webInterf import WebInterf, physBiblioWeb
except ImportError:
    print("Could not find physbiblio and its modules!")
    print(traceback.format_exc())
    raise


arxivCategories = {
    "astro-ph": ["CO", "EP", "GA", "HE", "IM", "SR"],
    "cond-mat": [
        "dis-nn",
        "mes-hall",
        "mtrl-sci",
        "other",
        "quant-gas",
        "soft",
        "stat-mech",
        "str-el",
        "supr-con",
    ],
    "cs": [
        "AI",
        "AR",
        "CC",
        "CE",
        "CG",
        "CL",
        "CR",
        "CV",
        "CY",
        "DB",
        "DC",
        "DL",
        "DM",
        "DS",
        "ET",
        "FL",
        "GL",
        "GR",
        "GT",
        "HC",
        "IR",
        "IT",
        "LG",
        "LO",
        "MA",
        "MM",
        "MS",
        "NA",
        "NE",
        "NI",
        "OH",
        "OS",
        "PF",
        "PL",
        "RO",
        "SC",
        "SD",
        "SE",
        "SI",
        "SY",
    ],
    "econ": ["EM", "GN", "TH"],
    "eess": ["AS", "IV", "SP"],
    "gr-qc": [],
    "hep-ex": [],
    "hep-lat": [],
    "hep-ph": [],
    "hep-th": [],
    "math": [
        "AC",
        "AG",
        "AP",
        "AT",
        "CA",
        "CO",
        "CT",
        "CV",
        "DG",
        "DS",
        "FA",
        "GM",
        "GN",
        "GR",
        "GT",
        "HO",
        "IT",
        "KT",
        "LO",
        "MG",
        "MP",
        "NA",
        "NT",
        "OA",
        "OC",
        "PR",
        "QA",
        "RA",
        "RT",
        "SG",
        "SP",
        "ST",
    ],
    "math-ph": [],
    "nlin": ["AO", "CD", "CG", "PS", "SI"],
    "nucl-ex": [],
    "nucl-th": [],
    "physics": [
        "acc-ph",
        "ao-ph",
        "app-ph",
        "atm-clus",
        "atom-ph",
        "bio-ph",
        "chem-ph",
        "class-ph",
        "comp-ph",
        "data-an",
        "ed-ph",
        "flu-dyn",
        "gen-ph",
        "geo-ph",
        "hist-ph",
        "ins-det",
        "med-ph",
        "optics",
        "plasm-ph",
        "pop-ph",
        "soc-ph",
        "space-ph",
    ],
    "q-bio": ["BM", "CB", "GN", "MN", "NC", "OT", "PE", "QM", "SC", "TO"],
    "q-fin": ["CP", "EC", "GN", "MF", "PM", "PR", "RM", "ST", "TR"],
    "quant-ph": [],
    "stat": ["AP", "CO", "ME", "ML", "OT", "TH"],
}


def isValidArxiv(string):
    """Determine if the string is a valid arXiv identifier
    (this does not mean that it exists)

    Parameter:
        string: the string to test

    Output:
        boolean
    """
    if "arxiv:" in string:
        string = string.replace("arxiv:", "")
    if re.match(
        "^[0-9]{4}\.[0-9]{4,5}(v[0-9]+)?$",
        string,
    ):
        return True
    elif re.match(
        "^[a-zA-Z]+(\-[a-zA-Z]+)?(\.[a-zA-Z]+(\-[a-zA-Z]+)?)?/[0-9]{7}(v[0-9]+)?$",
        string,
    ):
        if "." in string:
            matchcat = re.compile(
                "([a-zA-Z]+(\-[a-zA-Z]+)?)\.([a-zA-Z]+(\-[a-zA-Z]+)?)/"
            )
            for t in matchcat.finditer(string):
                cat = t.group(1)
                sub = t.group(3)
            if cat in arxivCategories and sub in arxivCategories[cat]:
                return True
        else:
            matchcat = re.compile("([a-zA-Z]+(\-[a-zA-Z]+)?)/")
            for t in matchcat.finditer(string):
                cat = t.group(1)
            if cat in arxivCategories:
                return True
        return False
    return False


def getYear(string):
    """Use the arxiv id to compute the year

    Parameter:
        string: the arxiv identifier

    Output:
        a string containing the year of the submission to arXiv
    """
    identif = re.compile("(([a-zA-Z\-]+/)([0-9]{7}))|([0-9]{4}\.[0-9]{4,5})")
    try:
        for t in identif.finditer(string):
            if len(t.group()) > 0:
                try:
                    a = t.group(3)[:2]  # match first type, e.g. hep-ex/0101001
                except (IndexError, TypeError):
                    a = t.group(4)[
                        :2
                    ]  # match second type, e.g. 1301.0123 or 1501.01234
                if int(a) > 90:
                    return "19" + a
                else:
                    return "20" + a
    except (IndexError, TypeError):
        pBLogger.warning(ArxivStrings.errorYearConversion % string)
        return None


class WebSearch(WebInterf, ArxivStrings):
    """Subclass of WebInterf that can connect to arxiv.org
    to perform searches
    """

    name = "arXiv"
    description = "arXiv fetcher"
    url = "https://export.arxiv.org/api/query"
    urlRss = "https://export.arxiv.org/rss/"
    categories = arxivCategories

    def __init__(self):
        """Initializes the class variables using
        the WebInterf constructor.

        Define additional specific parameters for the arxiv.org API.
        """
        WebInterf.__init__(self)
        self.urlArgs = {"start": "0"}

    def getYear(self, string):
        """Use the arxiv id to compute the year"""
        return getYear(string)

    def retrieveUrlFirst(self, string, searchType="all", **kwargs):
        """Retrieves the first result from the content
        of the given web page.
        The function calls arxivRetriever which will do the job.

        Parameters:
            string: the search string
            searchType: the search method in arxiv API (default 'all')

        Output:
            returns the bibtex string built from arxivRetriever
        """
        return self.arxivRetriever(
            string, searchType, additionalArgs={"max_results": "1"}, **kwargs
        )

    def retrieveUrlAll(self, string, searchType="all", **kwargs):
        """Retrieves all the results from the content
        of the given web page.
        The function calls arxivRetriever which will do the job.

        Parameters:
            string: the search string
            searchType: the search method in arxiv API (default 'all')

        Output:
            returns the bibtex string built from arxivRetriever
        """
        return self.arxivRetriever(
            string,
            searchType,
            additionalArgs={"max_results": pbConfig.params["maxExternalAPIResults"]},
            **kwargs
        )

    def arxivRetriever(
        self, string, searchType="all", additionalArgs=None, fullDict=False
    ):
        """Reads the feed content got from arxiv into a dictionary,
        used to return a bibtex.

        Parameters:
            string: the search string
            searchType: the search method in arxiv API (default 'all').
                The possible values are:
                    ti  -> Title
                    au  -> Author
                    abs -> Abstract
                    co  -> Comment
                    jr  -> Journal Reference
                    cat -> Subject Category
                    rn  -> Report Number
                    id  -> Id (use id_list instead)
                    all -> All of the above
            additionalArgs: a dictionary of additional arguments
                that can be passed to self.urlArgs (default None)
            fullDict (logical): return the bibtex dictionary in addition
                to the bibtex text (default False)

        Output:
            the bibtex text
            (optional, depending on fullDict): the bibtex Dictionary
        """

        def returnFailed():
            """Define output of the function when an error occurred"""
            if fullDict:
                return "", {}
            return ""

        urlArgs = self.urlArgs.copy()
        if additionalArgs:
            try:
                for k, v in additionalArgs.items():
                    urlArgs[k] = v
            except (AttributeError, TypeError):
                pass
        urlArgs["search_query"] = searchType + ":" + string
        url = self.createUrl(urlArgs)
        pBLogger.info(self.searchInfo % (searchType, string, url))
        text = parse_accents_str(self.textFromUrl(url))
        try:
            data = feedparser.parse(text)
        except Exception:  # intercept all possible errors
            pBLogger.exception(self.cannotParseRSS % text, exc_info=True)
            return returnFailed()
        if "entries" not in data.keys():
            pBLogger.exception(self.cannotFindEntriesInFeed % data)
            return returnFailed()
        db = BibDatabase()
        db.entries = []
        dictionaries = []
        for entry in data["entries"]:
            dictionary = {"ENTRYTYPE": "article", "archiveprefix": "arXiv"}
            try:
                idArx = (
                    entry["id"]
                    .replace("http://arxiv.org/abs/", "")
                    .replace("https://arxiv.org/abs/", "")
                )
            except (AttributeError, KeyError, TypeError):
                pBLogger.exception(self.cannotReadEntryId % entry)
                continue
            else:
                pos = idArx.find("v")
                if pos >= 0:
                    idArx = idArx[0:pos]
                dictionary["ID"] = idArx
                dictionary["arxiv"] = idArx
            for m, k in (("title", "title"), ("doi", "arxiv_doi")):
                try:
                    dictionary[m] = entry[k]
                except KeyError as e:
                    pBLogger.debug("KeyError: %s" % e)
            try:
                dictionary["abstract"] = entry["summary"].replace("\n", " ")
            except KeyError as e:
                pBLogger.debug("KeyError: %s" % e)
            try:
                dictionary["authors"] = " and ".join(
                    [au["name"] for au in entry["authors"]]
                )
            except KeyError as e:
                pBLogger.debug("KeyError: %s" % e)
            try:
                dictionary["primaryclass"] = entry["arxiv_primary_category"]["term"]
            except KeyError as e:
                pBLogger.debug("KeyError: %s" % e)
            try:
                year = self.getYear(dictionary["arxiv"])
            except KeyError as e:
                pBLogger.debug("KeyError: %s" % e)
            else:
                if year is not None:
                    dictionary["year"] = year
            db.entries.append(dictionary)
            dictionaries.append(dictionary)
        writtendb = pbWriter.write(db)
        if fullDict:
            try:
                dictionary = dictionaries[0]
                for d in dictionaries:
                    if string in d["arxiv"]:
                        dictionary = d
            except (IndexError, TypeError):
                pBLogger.exception(self.cannotReadDict)
                return returnFailed()
            else:
                return writtendb, dictionary
        else:
            return writtendb

    def arxivDaily(self, category):
        """Read daily RSS feed for a given category

        Parameter:
            category: the selected category (see `self.categories)
        """
        if "." in category:
            main, sub = category.split(".")
        else:
            main = category
            sub = ""
        url = self.urlRss
        if main not in self.categories.keys():
            pBLogger.warning(self.mainCatNotFound % main)
            return False
        else:
            url += main
        if sub != "" and sub not in self.categories[main]:
            pBLogger.warning(self.subCatNotFound % sub)
            return False
        elif sub != "" and sub in self.categories[main]:
            url += "." + sub
        pBLogger.info(url)
        text = self.textFromUrl(url)
        if text is None:
            pBLogger.warning(self.emptyUrl)
            return False
        author = re.compile("(>|&gt;)([^/]*)(</a>|&lt;/a&gt;)")
        additionalInfo = re.compile(
            " \(arXiv:([0-9\.v]*) \[([\-\.a-zA-Z]*)\]([ A-Z]*)\)"
        )
        text = parse_accents_str(text)
        try:
            data = feedparser.parse(text)
        except Exception:
            pBLogger.exception(self.cannotParseRSS % text)
            return False
        entries = []
        for element in data.entries:
            tmp = {}
            try:
                tmp["eprint"] = element["id"].split("/")[-1]
                tmp["abstract"] = (
                    element["summary"]
                    .replace("\n", " ")
                    .replace("<p>", "")
                    .replace("</p>", "")
                )
                tmp["authors"] = [
                    m.group(2)
                    for m in author.finditer(element["authors"][0]["name"])
                    if m != ""
                ]
                tmp["author"] = (
                    " and ".join(tmp["authors"])
                    if len(tmp["authors"]) < pbConfig.params["maxAuthorNames"]
                    else " and ".join(
                        tmp["authors"][0 : pbConfig.params["maxAuthorNames"]]
                        + ["others"]
                    )
                )
                tmp["replacement"] = "UPDATED" in element["title"]
                tmp["primaryclass"] = [
                    m.group(2)
                    for m in additionalInfo.finditer(element["title"])
                    if m != ""
                ][0]
                tmp["cross"] = (
                    "CROSS LISTED" in element["title"]
                    or category.lower() not in tmp["primaryclass"].lower()
                )
                tmp["version"] = [
                    m.group(1)
                    for m in additionalInfo.finditer(element["title"])
                    if m != ""
                ][0]
                parenthesis = [
                    m.group()
                    for m in additionalInfo.finditer(element["title"])
                    if m != ""
                ][0]
                tmp["title"] = element["title"].replace(parenthesis, "")
            except (IndexError, KeyError, TypeError):
                pBLogger.warning(self.cannotReadItem)
                continue
            else:
                entries.append(tmp)
        return entries



"""Module that manages the actions on the database (and few more).

This file is part of the physbiblio package.
"""
import ast
import os
import sqlite3
import sys
import traceback
from sqlite3 import (
    DatabaseError,
    IntegrityError,
    InterfaceError,
    OperationalError,
    ProgrammingError,
)

try:
    import physbiblio.tablesDef
    from physbiblio.strings.main import DatabaseCoreStrings as dbcstr
except ImportError:
    print("Could not find physbiblio and its modules!")
    print(traceback.format_exc())
    raise

encoding_default = "iso-8859-15"


class PhysBiblioDBCore:
    """Contains most of the basic functions on the database.
    Will be subclassed to do everything else.
    """

    def __init__(self, dbname, logger, noOpen=False, info=True):
        """Initialize database class (column names, descriptions)
        and opens the database.

        Parameters:
            dbname: the name of the database to be opened
            logger: the logger used for messages
            noOpen (boolean, default False): open the database or not
            info (boolean, default True): show some output when opening DB
        """
        # structure of the tables
        self.tableFields = physbiblio.tablesDef.tableFields
        self.descriptions = physbiblio.tablesDef.fieldsDescriptions
        # names of the columns
        self.tableCols = {}
        for q in self.tableFields.keys():
            self.tableCols[q] = [a[0] for a in self.tableFields[q]]

        self.dbChanged = False
        self.conn = None
        self.curs = None
        self.onIsLocked = None
        self.dbname = dbname
        self.logger = logger
        db_is_new = not os.path.exists(self.dbname)

        if not noOpen:
            self.openDB(info=info)
            if db_is_new or self.checkExistingTables():
                self.logger.info(dbcstr.noDatabaseCreate)
                self.createTables()
            self.checkDatabaseUpdates()

        self.lastFetched = None
        self.catsHier = None

        self.loadSubClasses()

    def openDB(self, info=True):
        """Open the database and creates the self.conn (connection)
        and self.curs (cursor) objects.

        Parameters:
            info (boolean, default True): show some output when opening DB

        Output:
            True
        """
        if info:
            self.logger.info(dbcstr.openDb % self.dbname)
        else:
            self.logger.debug(dbcstr.openDb % self.dbname)
        self.conn = sqlite3.connect(self.dbname, check_same_thread=False)
        self.conn.row_factory = sqlite3.Row
        self.curs = self.conn.cursor()
        self.loadSubClasses()
        return True

    def reOpenDB(self):
        """Not defined at this stage. Present in subclass PhysBiblioDB"""
        pass

    def closeDB(self, info=True):
        """Close the database.

        Parameters:
            info (boolean, default True): show some output when opening DB
        """
        if info:
            self.logger.info(dbcstr.closeDb)
        else:
            self.logger.debug(dbcstr.closeDb)
        self.conn.close()
        return True

    def sendDBIsLocked(self):
        """If the database is open in another instance
        and no writing operation are allowed,
        try to emit a signal to the main GUI
        """
        if self.onIsLocked is not None:
            try:
                self.onIsLocked.emit()
            except AttributeError:
                self.logger.exception(dbcstr.invalidIsLocked)
                return False
            else:
                return True
        return False

    def checkUncommitted(self):
        """Check if there are uncommitted changes.

        Output:
            True/False
        """
        return self.conn.in_transaction

    def commit(self, verbose=True):
        """Commit the changes.

        Parameters:
            verbose (boolean, default True):
                show some output when opening DB

        Output:
            True if successfull, False if an exception occurred
        """
        try:
            self.conn.commit()
        except Exception:
            self.logger.exception(dbcstr.errorCannotCommit)
            return False
        else:
            self.dbChanged = False
            if verbose:
                self.logger.info(dbcstr.savedDb)
            return True

    def undo(self, verbose=True):
        """Undo the uncommitted changes
        and roll back to the last commit.

        Parameters:
            verbose (boolean, default True):
                show some output when opening DB

        Output:
            True if successfull, False if an exception occurred
        """
        try:
            self.conn.rollback()
        except Exception:
            self.logger.exception(dbcstr.errorCannotRollback)
            return False
        else:
            self.dbChanged = False
            if verbose:
                self.logger.info(dbcstr.rollbackDb)
            return True

    def connExec(self, query, data=None):
        """Execute connection.

        Parameters:
            query (string): the query to be executed
            data (dictionary or list):
                the values of the parameters in the query

        Output:
            True if successfull, False if an exception occurred
        """
        try:
            if data:
                self.conn.execute(query, data)
            else:
                self.conn.execute(query)
        except OperationalError as err:
            if str(err) == "database is locked":
                if not self.sendDBIsLocked():
                    self.logger.error(dbcstr.opErDbOpen % query)
            else:
                self.logger.exception(dbcstr.errorConnection % (err, query))
            return False
        except IntegrityError as err:
            self.logger.exception(dbcstr.errorInsUpd % (err, query))
            return False
        except (ProgrammingError, DatabaseError, InterfaceError) as err:
            self.logger.exception(dbcstr.errorConnection % (err, query))
            return False
        else:
            self.dbChanged = True
            return True

    def cursExec(self, query, data=None):
        """Execute cursor.

        Parameters:
            query (string): the query to be executed
            data (dictionary or list):
                the values of the parameters in the query

        Output:
            True if successfull, False if an exception occurred
        """
        try:
            if data:
                self.curs.execute(query, data)
            else:
                self.curs.execute(query)
        except Exception as err:
            self.logger.exception(
                dbcstr.errorCursor
                % (
                    err,
                    query,
                    data,
                )
            )
            return False
        else:
            return True

    def cursor(self):
        """Function wrapper that returns the default cursor"""
        return self.curs

    def loadSubClasses(self):
        """Not defined at this stage. Present in subclass PhysBiblioDB"""
        pass

    def checkExistingTables(
        self,
        wantedTables=[
            "categories",
            "entries",
            "entryCats",
            "entryExps",
            "expCats",
            "experiments",
            "settings",
        ],
    ):
        """Check that all the required tables
        are present in the database
        """
        self.cursExec("SELECT name FROM sqlite_master WHERE type='table';")
        tables = [name[0] for name in self.curs]
        return not all(t in tables for t in wantedTables)

    def createTable(self, q, fields, critical=False):
        """Create the table 'q'

        Parameters:
            q: the table name
            fieldsDict: the list containing the column information
                for the table
        """
        command = "CREATE TABLE %s (\n" % q
        first = True
        for el in fields:
            if first:
                first = False
            else:
                command += ",\n"
            command += " ".join(el)
        command += ");"
        self.logger.info(command + "\n")
        if not self.connExec(command):
            self.logger.critical(dbcstr.errorCreateTable % q)
            if critical:
                sys.exit(1)
        else:
            if critical:
                self.commit()

    def createTables(self, fieldsDict=None):
        """Create tables for the database
        (and insert the default categories), if it is missing.

        Parameters:
            fieldsDict (default None):
                the structure of the tables (see physbiblio.tablesDef)
        """
        if fieldsDict is None:
            fieldsDict = self.tableFields
        self.cursExec("SELECT name FROM sqlite_master WHERE type='table';")
        existingTables = [name[0] for name in self.curs]
        for q in fieldsDict.keys():
            if q in existingTables:
                continue
            self.createTable(q, fieldsDict[q])
        self.cursExec("select * from categories where idCat = 0 or idCat = 1\n")
        cats = self.curs.fetchall()
        if len(cats) < 2:
            command = (
                "INSERT into categories "
                + "(idCat, name, description, parentCat, ord) values "
                + '(0,"Main","This is the main category. '
                + 'All the other ones are subcategories of this one",0,0), '
                + '(1,"Tags","Use this category to store tags '
                + '(such as: ongoing projects, temporary cats,...)",0,0)\n'
            )
            self.logger.info(command + "\n")
            if not self.connExec(command):
                self.logger.error(dbcstr.errorInsMainCats)
        self.commit()

    def checkDatabaseUpdates(self):
        """Run when new columns are added to the database with respect
        to previous versions of the software
        Check if bibdict column is present in entries table
        """
        entryFields = self.cursExec("PRAGMA table_info(entries);")
        entriesCols = [name[1] for name in self.curs]
        if "bibdict" not in entriesCols:
            if self.connExec("ALTER TABLE entries ADD COLUMN bibdict text;"):
                self.logger.info(dbcstr.newColEntries % ("bibdict", "text"))
                self.commit()
            else:
                self.logger.error(dbcstr.errorAlterEntries)
                self.undo()
        if "citations" not in entriesCols:
            if self.connExec(
                "ALTER TABLE entries ADD COLUMN citations integer default 0;"
            ):
                self.logger.info(dbcstr.newColEntries % ("citations", "integer"))
                self.commit()
            else:
                self.logger.error(dbcstr.errorAlterEntries)
                self.undo()
            if self.connExec(
                "ALTER TABLE entries ADD COLUMN citations_no_self integer default 0;"
            ):
                self.logger.info(
                    dbcstr.newColEntries % ("citations_no_self", "integer")
                )
                self.commit()
            else:
                self.logger.error(dbcstr.errorAlterEntries)
                self.undo()


class PhysBiblioDBSub:
    """Uses PhysBiblioDB instance 'self.mainDB = parent'
    to act on the database.
    All the subcategories of PhysBiblioDB are defined
    starting from this one.
    """

    def __init__(self, parent):
        """Initialize DB class, connecting to
        the main PhysBiblioDB instance (parent).
        """
        self.mainDB = parent

        self.tableFields = self.mainDB.tableFields
        self.tableCols = self.mainDB.tableCols

        self.conn = self.mainDB.conn
        self.curs = self.mainDB.curs
        self.dbname = self.mainDB.dbname

        self.lastFetched = None
        self.catsHier = None

    def sendDBIsLocked(self):
        """Call corresponding PhysBiblioDBCore function"""
        return self.mainDB.sendDBIsLocked()

    def literal_eval(self, string):
        """Wrapper for ast.literal_eval

        Parameters:
            string: the string to evaluate

        Output:
            a string or None
        """
        try:
            if "[" in string and "]" in string:
                return ast.literal_eval(string.strip())
            elif "," in string:
                return ast.literal_eval("[%s]" % string.strip())
            else:
                return string.strip()
        except SyntaxError:
            self.mainDB.logger.warning(dbcstr.errorLiteralEval % string)
            return None

    def closeDB(self):
        """Close the database (using PhysBiblioDB.close)"""
        self.mainDB.closeDB()

    def commit(self):
        """Commit the changes (using PhysBiblioDB.commit)"""
        self.mainDB.commit()

    def connExec(self, query, data=None):
        """Execute connection (see PhysBiblioDB.connExec)"""
        return self.mainDB.connExec(query, data=data)

    def cursExec(self, query, data=None):
        """Execute cursor (see PhysBiblioDB.cursExec)"""
        return self.mainDB.cursExec(query, data=data)

    def cursor(self):
        """Return the cursor"""
        return self.mainDB.cursor()



#!/usr/bin/env python
"""Test file for the physbiblio.webimport subpackage.

This file is part of the physbiblio package.
"""
import datetime
import json
import os
import traceback
import unittest
from unittest.mock import MagicMock, patch

from ads.tests.stubdata.export import example_export_response
from ads.tests.stubdata.solr import example_solr_response

try:
    from physbiblio.config import pbConfig
    from physbiblio.setuptests import *
    from physbiblio.webimport.adsnasa import WebSearch
    from physbiblio.webimport.webInterf import WebInterf, physBiblioWeb
except ImportError:
    print("Could not find physbiblio and its modules!")
    raise
except Exception:
    print(traceback.format_exc())


class TestADSOnlineMethods(unittest.TestCase):
    """Test the functions that import entries from the ADS.
    Should not fail if everything works fine.
    """

    def test_init(self):
        """Test init and basic class properties"""
        ws = WebSearch()
        self.assertIsInstance(ws, WebInterf)
        self.assertTrue(hasattr(ws, "name"))
        self.assertTrue(hasattr(ws, "url"))
        self.assertTrue(hasattr(ws, "loadFields"))
        self.assertTrue(hasattr(ws, "fewFields"))
        self.assertIsInstance(physBiblioWeb.webSearch["adsnasa"], WebSearch)

    @unittest.skipIf(skipTestsSettings.online, "Online tests")
    def test_getGenericInfo(self):
        """Test getGenericInfo"""
        with patch("ads.SearchQuery", return_value=("abc", "def")) as _sq, patch(
            "logging.Logger.info"
        ) as _in, patch("logging.Logger.exception") as _ex, patch(
            "physbiblio.webimport.adsnasa.WebSearch.getLimitInfo",
            autospec=True,
            return_value="nolimit",
        ) as _gl:
            a = physBiblioWeb.webSearch["adsnasa"].getGenericInfo("star", ["bibcode"])
            self.assertEqual(a, ["abc", "def"])
            _sq.assert_called_once_with(
                q="star", fl=["bibcode"], rows=pbConfig.params["maxExternalAPIResults"]
            )
            self.assertEqual(_in.call_count, 1)
            self.assertEqual(_gl.call_count, 1)
            self.assertEqual(_ex.call_count, 0)
        with patch.dict(pbConfig.params, {"ADSToken": [""]}, clear=False), patch(
            "logging.Logger.exception"
        ) as _ex:
            a = physBiblioWeb.webSearch["adsnasa"].getGenericInfo("star", ["bibcode"])
            self.assertEqual(a, [])
            _ex.assert_called_once_with(
                "Unauthorized use of ADS API. Is your token valid?"
            )

    @unittest.skipIf(skipTestsSettings.online, "Online tests")
    def test_getBibtexs(self):
        """Test getBibtexs"""
        fr = MagicMock()
        fr.execute = MagicMock(return_value="exported")
        with patch("ads.ExportQuery", return_value=fr) as _eq, patch(
            "logging.Logger.info"
        ) as _in, patch("logging.Logger.exception") as _ex, patch(
            "physbiblio.webimport.adsnasa.WebSearch.getLimitInfo",
            autospec=True,
            return_value="nolimit",
        ) as _gl:
            a = physBiblioWeb.webSearch["adsnasa"].getBibtexs(["bibcode"])
            _eq.assert_called_once_with(bibcodes=["bibcode"], format="bibtex")
            _in.assert_called_once_with("nolimit")
            _gl.assert_called_once_with(physBiblioWeb.webSearch["adsnasa"])
            self.assertEqual(_ex.call_count, 0)
            fr.execute.assert_called_once_with()
            self.assertEqual(a, "exported")
        with patch.dict(pbConfig.params, {"ADSToken": [""]}, clear=False), patch(
            "logging.Logger.exception"
        ) as _ex:
            a = physBiblioWeb.webSearch["adsnasa"].getBibtexs(["bibcode"])
            self.assertEqual(a, "")
            _ex.assert_called_once_with(
                "Unauthorized use of ADS API. Is your token valid?"
            )

    def test_retrieveUrl(self):
        """Test retrieveUrlFirst and retrieveUrlAll"""
        a1 = MagicMock()
        a1.bibcode = "a1"
        a2 = MagicMock()
        a2.bibcode = "a2"
        with patch(
            "physbiblio.webimport.adsnasa.WebSearch.getGenericInfo",
            autospec=True,
            return_value=[a1, a2],
        ) as _ggi, patch(
            "physbiblio.webimport.adsnasa.WebSearch.getBibtexs",
            autospec=True,
            return_value="bibtex",
        ) as _gb:
            a = physBiblioWeb.webSearch["adsnasa"].retrieveUrlFirst("star")
            _ggi.assert_called_once_with(
                physBiblioWeb.webSearch["adsnasa"],
                "star",
                physBiblioWeb.webSearch["adsnasa"].fewFields,
            )
            _gb.assert_called_once_with(physBiblioWeb.webSearch["adsnasa"], "a1")
            self.assertEqual(a, "bibtex")
            _ggi.reset_mock()
            _gb.reset_mock()
            a = physBiblioWeb.webSearch["adsnasa"].retrieveUrlAll("star")
            _ggi.assert_called_once_with(
                physBiblioWeb.webSearch["adsnasa"],
                "star",
                physBiblioWeb.webSearch["adsnasa"].fewFields,
            )
            _gb.assert_called_once_with(
                physBiblioWeb.webSearch["adsnasa"], ["a1", "a2"]
            )
            self.assertEqual(a, "bibtex")
        with patch(
            "physbiblio.webimport.adsnasa.WebSearch.getGenericInfo",
            autospec=True,
            return_value=[],
        ) as _ggi, patch(
            "physbiblio.webimport.adsnasa.WebSearch.getBibtexs",
            autospec=True,
            return_value="abc",
        ) as _gb:
            a = physBiblioWeb.webSearch["adsnasa"].retrieveUrlFirst("star")
            self.assertEqual(a, "")
            a = physBiblioWeb.webSearch["adsnasa"].retrieveUrlAll("star")
            self.assertEqual(a, "abc")

    def test_getField(self):
        """Test getField"""
        pa = physBiblioWeb.webSearch["adsnasa"]
        a1 = MagicMock()
        a1.bibcode = "a1"
        a2 = MagicMock()
        a2.bibcode = "a2"
        with patch(
            "physbiblio.webimport.adsnasa.WebSearch.getGenericInfo",
            autospec=True,
            return_value=[a1, a2],
        ) as _ggi, patch("logging.Logger.warning") as _w:
            a = pa.getField("star", "bibcode")
            _ggi.assert_called_once_with(pa, "star", pa.loadFields)
            self.assertEqual(a, ["a1", "a2"])
            self.assertEqual(_w.call_count, 0)
            a = pa.getField("star", "somefield")
            _ggi.assert_called_once_with(pa, "star", pa.loadFields)
            self.assertEqual(a, [])
            _w.assert_called_once_with(
                "Invalid requested field in ADS query: somefield"
            )

    def test_getLimitInfo(self):
        """Test getLimitInfo"""
        pa = physBiblioWeb.webSearch["adsnasa"]
        pa.q = MagicMock()
        pa.q.response = MagicMock()
        pa.q.response.get_ratelimits = MagicMock(
            return_value={"limit": "a", "remaining": "123"}
        )
        self.assertEqual(pa.getLimitInfo(), "Remaining queries: 123/a")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""Test file for the physbiblio.pdf module.

This file is part of the physbiblio package.
"""
import os
import shutil
import traceback
import unittest
from unittest.mock import call, patch

try:
    from physbiblio.config import pbConfig
    from physbiblio.database import PhysBiblioDB, pBDB
    from physbiblio.pdf import LocalPDF, pBPDF
    from physbiblio.setuptests import *
except ImportError:
    print("Could not find physbiblio and its modules!")
    raise
except Exception:
    print(traceback.format_exc())


pBPDF.pdfDir = os.path.join(pbConfig.dataPath, "testpdf_%s" % today_ymd)


@unittest.skipIf(skipTestsSettings.long, "Long tests")
@patch("logging.Logger.debug")
@patch("logging.Logger.info")
@patch("logging.Logger.warning")
@patch("logging.Logger.exception")
class TestPdfMethods(unittest.TestCase):
    """Test the methods and functions in the pdf module"""

    def test_init(self, *args):
        """Test the __init__ method"""
        with patch(
            "physbiblio.pdf.LocalPDF.checkFolderExists", autospec=True
        ) as _cf, patch.dict(
            pbConfig.params,
            {"pdfFolder": "/a/b/c", "pdfApplication": "someapp"},
            clear=False,
        ):
            pdf = LocalPDF()
            self.assertEqual(pdf.pdfDir, "/a/b/c")
            self.assertEqual(pdf.pdfApp, "someapp")
            self.assertEqual(pdf.badFNameCharacters, "\\/:*?\"<>|'")
            _cf.assert_called_once_with(pdf)

    def test_checkFolderExists(self, *args):
        """Test the checkFolderExists method"""
        with patch(
            "physbiblio.pdf.LocalPDF.checkFolderExists", autospec=True
        ) as _cf, patch.dict(
            pbConfig.params,
            {"pdfFolder": "/a/b/c", "pdfApplication": "someapp"},
            clear=False,
        ):
            pdf = LocalPDF()
        with patch(
            "os.path.isdir", side_effect=[True, False], autospec=True
        ) as _id, patch("logging.Logger.info") as _if, patch("os.makedirs") as _mk:
            pdf.checkFolderExists()
            _id.assert_called_once_with("/a/b/c")
            _if.assert_not_called()
            _mk.assert_not_called()
            pdf.checkFolderExists()
            _if.assert_called_once_with("PDF folder is missing: /a/b/c. Creating it.")
            _mk.assert_called_once_with("/a/b/c")

    def test_fnames(self, *args):
        """Test names of folders and directories"""
        self.assertEqual(
            pBPDF.badFName(r'a\b/c:d*e?f"g<h>i|' + "j'"), "a_b_c_d_e_f_g_h_i_j_"
        )
        self.assertEqual(pBPDF.badFName(True), "")
        self.assertEqual(
            pBPDF.getFileDir(r'a\b/c:d*e?f"g<h>i|' + "j'"),
            os.path.join(pBPDF.pdfDir, "a_b_c_d_e_f_g_h_i_j_"),
        )
        testPaperFolder = pBPDF.getFileDir("abc.def")
        with patch(
            "physbiblio.database.Entries.getField",
            return_value="12345678",
            autospec=True,
        ) as _mock:
            self.assertEqual(
                pBPDF.getFilePath("abc.def", "arxiv"),
                os.path.join(testPaperFolder, "12345678.pdf"),
            )

    def test_manageFiles(self, *args):
        """Test creation, copy and deletion of files and folders"""
        emptyPdfName = os.path.join(pBPDF.pdfDir, "tests_%s.pdf" % today_ymd)
        pBPDF.createFolder("abc.def")
        self.assertTrue(os.path.exists(pBPDF.getFileDir("abc.def")))
        pBPDF.renameFolder("abc.def", "abc.fed")
        self.assertFalse(os.path.exists(pBPDF.getFileDir("abc.def")))
        self.assertTrue(os.path.exists(pBPDF.getFileDir("abc.fed")))
        pBPDF.createFolder("abc.def")
        open(
            os.path.join(pBPDF.pdfDir, "abc.def", "tests_%s.pdf" % today_ymd), "a"
        ).close()
        pBPDF.renameFolder("abc.def", "abc.fed")
        self.assertFalse(os.path.exists(pBPDF.getFileDir("abc.def")))
        self.assertTrue(os.path.exists(pBPDF.getFileDir("abc.fed")))
        self.assertFalse(
            os.path.exists(pBPDF.getFileDir("abd.fed" + os.sep + "abc.def"))
        )
        self.assertTrue(
            os.path.exists(
                os.path.join(pBPDF.pdfDir, "abc.fed", "tests_%s.pdf" % today_ymd)
            )
        )
        open(emptyPdfName, "a").close()
        with patch(
            "physbiblio.database.Entries.getField",
            return_value="12345678",
            autospec=True,
        ) as _mock:
            pBPDF.copyNewFile("abc.fed", emptyPdfName, "arxiv")
            pBPDF.copyNewFile("abc.fed", emptyPdfName, customName="empty.pdf")
            self.assertTrue(os.path.exists(pBPDF.getFilePath("abc.fed", "arxiv")))
            os.remove(emptyPdfName)
            self.assertFalse(os.path.exists(emptyPdfName))
            pBPDF.copyToDir(pBPDF.pdfDir, "abc.fed", "arxiv")
            self.assertTrue(os.path.exists(os.path.join(pBPDF.pdfDir, "12345678.pdf")))
            pBPDF.mergePDFFolders("abc.fed", "cba.fed")
            self.assertTrue(os.path.exists(pBPDF.getFilePath("cba.fed", "arxiv")))
            self.assertTrue(pBPDF.removeFile("abc.fed", "arxiv"))
            self.assertFalse(pBPDF.removeFile("abc.fed", "arxiv"))
            os.remove(os.path.join(pBPDF.pdfDir, "12345678.pdf"))
            shutil.rmtree(pBPDF.getFileDir("abc.fed"))
            shutil.rmtree(pBPDF.getFileDir("cba.fed"))

    @unittest.skipIf(skipTestsSettings.online, "Online tests")
    def test_download(self, *args):
        """Test downloadArxiv"""
        with patch(
            "physbiblio.database.Entries.getField",
            return_value="1806.11344",
            autospec=True,
        ) as _mock:
            pBPDF.createFolder("abc.def")
            with open(pBPDF.getFilePath("abc.def", "arxiv"), "w") as _fe:
                _fe.write("a")
            self.assertTrue(pBPDF.downloadArxiv("abc.def"))
            self.assertTrue(pBPDF.checkFile("abc.def", "arxiv"))
            self.assertEqual(pBPDF.getExisting("abc.def"), ["1806.11344.pdf"])
            self.assertEqual(
                pBPDF.getExisting("abc.def", fullPath=True),
                [os.path.join(pBPDF.getFileDir("abc.def"), "1806.11344.pdf")],
            )
            open(os.path.join(pBPDF.getFileDir("abc.def"), "1806.11344"), "w").close()
            self.assertEqual(
                sorted(pBPDF.getExisting("abc.def")),
                sorted(["1806.11344", "1806.11344.pdf"]),
            )
            self.assertTrue(
                pBPDF.removeFile(
                    "abc.def",
                    "file",
                    os.path.join(pBPDF.getFileDir("abc.def"), "1806.11344"),
                )
            )
            self.assertEqual(pBPDF.getExisting("abc.def"), ["1806.11344.pdf"])
            self.assertTrue(pBPDF.removeFile("abc.def", "arxiv"))
            self.assertFalse(pBPDF.checkFile("abc.def", "arxiv"))
        with patch(
            "physbiblio.database.Entries.getField",
            side_effect=["1801.15000", "1801.15000", "", "", None, None],
            autospec=True,
        ) as _mock:
            self.assertFalse(pBPDF.downloadArxiv("abc.def"))
            self.assertFalse(pBPDF.downloadArxiv("abc.def"))
            self.assertFalse(pBPDF.downloadArxiv("abc.def"))
        shutil.rmtree(pBPDF.getFileDir("abc.def"))

    def test_removeSpare(self, *args):
        """Test finding spare folders"""
        with patch(
            "physbiblio.database.Entries.fetchCursor",
            return_value=[{"bibkey": "abc"}, {"bibkey": "def"}],
            autospec=True,
        ) as _mock:
            for q in ("abc", "def", "ghi"):
                pBPDF.createFolder(q)
                self.assertTrue(os.path.exists(pBPDF.getFileDir(q)))
            pBPDF.removeSparePDFFolders()
            for q in ("abc", "def"):
                self.assertTrue(os.path.exists(pBPDF.getFileDir(q)))
            self.assertFalse(os.path.exists(pBPDF.getFileDir("ghi")))
        shutil.rmtree(pBPDF.pdfDir)

    def test_numberOfFiles(self, *args):
        """test numberOfFiles"""
        self.assertEqual(pBPDF.numberOfFiles("/surely/non/existent/folder"), 0)
        with patch(
            "os.walk",
            return_value=[["f", ["b"], ["a", "b/c", "b/d", "b/e"]]],
            autospec=True,
        ) as _wa, patch(
            "os.path.isfile", side_effect=[True, True, True, False], autospec=True
        ) as _if:
            self.assertEqual(pBPDF.numberOfFiles("f"), 3)
            _wa.assert_called_once_with("f")
            _if.assert_has_calls(
                [call("f/a"), call("f/b/c"), call("f/b/d"), call("f/b/e")]
            )

    def test_dirSize(self, *args):
        """test dirSize"""
        error_class = FileNotFoundError
        with patch("logging.Logger.exception") as _e, patch(
            "os.makedirs"
        ) as _md, patch("os.path.getsize", side_effect=[error_class, 123]) as _gs:
            self.assertEqual(pBPDF.dirSize("/surely/non/existent/folder"), 123)
            _e.assert_called_once_with(
                "PDF folder is missing: /surely/non/existent/folder. Creating it."
            )
            _md.assert_called_once_with("/surely/non/existent/folder")
            _gs.assert_any_call("/surely/non/existent/folder")
        with patch(
            "os.path.getsize", side_effect=[1, 1, 100, 100, 100], autospec=True
        ) as _gs, patch(
            "os.walk",
            return_value=[["f", ["b"], ["a", "b/c", "b/d", "b/e"]]],
            autospec=True,
        ) as _wa, patch(
            "os.path.isfile", side_effect=[True, True, True, False], autospec=True
        ) as _if:
            self.assertEqual(pBPDF.dirSize("f"), 302)
            _gs.assert_has_calls(
                [call("f"), call("f/b"), call("f/a"), call("f/b/c"), call("f/b/d")]
            )
            _wa.assert_called_once_with("f")
            _if.assert_has_calls(
                [call("f/a"), call("f/b/c"), call("f/b/d"), call("f/b/e")]
            )
        with patch(
            "os.path.getsize", side_effect=[100, 100, 100], autospec=True
        ) as _gs, patch(
            "os.walk",
            return_value=[["f", ["b"], ["a", "b/c", "b/d", "b/e"]]],
            autospec=True,
        ) as _wa, patch(
            "os.path.isfile", side_effect=[True, True, True, False], autospec=True
        ) as _if:
            self.assertEqual(pBPDF.dirSize("f", dirs=False), 300)
            _gs.assert_has_calls([call("f/a"), call("f/b/c"), call("f/b/d")])
            _wa.assert_called_once_with("f")
            _if.assert_has_calls(
                [call("f/a"), call("f/b/c"), call("f/b/d"), call("f/b/e")]
            )

    def test_getSizeWUnits(self, *args):
        """test getSizeWUnits"""
        with patch("logging.Logger.warning") as _w:
            self.assertEqual(pBPDF.getSizeWUnits(2048**2), "4.00MB")
            self.assertEqual(pBPDF.getSizeWUnits(2048**2, units="mb"), "4.00MB")
            self.assertEqual(pBPDF.getSizeWUnits(2048**2, units="kb"), "4096.00KB")
            self.assertEqual(pBPDF.getSizeWUnits(2048**2, units="PB"), "0.00PB")
            self.assertEqual(
                pBPDF.getSizeWUnits(2048**2, units="gb", fmt="%.3f"), "0.004GB"
            )
            _w.assert_not_called()
        with patch("logging.Logger.warning") as _w:
            self.assertEqual(pBPDF.getSizeWUnits(2048**2, units="m"), "4.00MB")
            _w.assert_called_once_with("Invalid units. Changing to 'MB'.")
        with patch("logging.Logger.warning") as _w:
            self.assertEqual(pBPDF.getSizeWUnits("a"), "nan")
            _w.assert_called_once_with("Invalid size. It must be a number!")
        with patch("logging.Logger.warning") as _w:
            self.assertEqual(pBPDF.getSizeWUnits(2048**2, fmt="%"), "4.00MB")
            _w.assert_called_once_with("Invalid format. Using '%.2f'")


if __name__ == "__main__":
    unittest.main()
